# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.definitions import QUANTIZATION_ALGO_NAME_IN_CONFIG
from nncf.config.schemata.basic import ARRAY_OF_NUMBERS
from nncf.config.schemata.basic import ARRAY_OF_STRINGS
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import annotated_enum
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.common.compression import COMPRESSION_LR_MULTIPLIER_PROPERTY
from nncf.config.schemata.common.initialization import BATCHNORM_ADAPTATION_SCHEMA
from nncf.config.schemata.common.targeting import SCOPING_PROPERTIES
from nncf.config.schemata.defaults import ACTIVATIONS_QUANT_START_EPOCH
from nncf.config.schemata.defaults import AUTOQ_EVAL_SUBSET_RATIO
from nncf.config.schemata.defaults import AUTOQ_WARMUP_ITER_NUMBER
from nncf.config.schemata.defaults import HAWQ_DUMP_INIT_PRECISION_DATA
from nncf.config.schemata.defaults import HAWQ_ITER_NUMBER
from nncf.config.schemata.defaults import HAWQ_NUM_DATA_POINTS
from nncf.config.schemata.defaults import HAWQ_TOLERANCE
from nncf.config.schemata.defaults import LR_POLY_DURATION_EPOCHS
from nncf.config.schemata.defaults import MAX_PERCENTILE
from nncf.config.schemata.defaults import MIN_PERCENTILE
from nncf.config.schemata.defaults import NUM_INIT_SAMPLES
from nncf.config.schemata.defaults import PRECISION_INIT_BITWIDTHS
from nncf.config.schemata.defaults import QUANTIZATION_BITS
from nncf.config.schemata.defaults import QUANTIZATION_EXPORT_TO_ONNX_STANDARD_OPS
from nncf.config.schemata.defaults import QUANTIZATION_LOGARITHM_SCALE
from nncf.config.schemata.defaults import QUANTIZATION_OVERFLOW_FIX
from nncf.config.schemata.defaults import QUANTIZATION_PER_CHANNEL
from nncf.config.schemata.defaults import QUANTIZATION_PRESET
from nncf.config.schemata.defaults import QUANTIZE_INPUTS
from nncf.config.schemata.defaults import QUANTIZE_OUTPUTS
from nncf.config.schemata.defaults import STAGED_QUANTIZATION_BASE_LR
from nncf.config.schemata.defaults import STAGED_QUANTIZATION_BASE_WD
from nncf.config.schemata.defaults import WEIGHTS_QUANT_START_EPOCH

QUANTIZER_MODES = ["symmetric", "asymmetric"]

QUANTIZER_CONFIG_PROPERTIES = {
    "mode": with_attributes(
        STRING,
        description=f"Mode of quantization. "
        f"See [Quantization.md]"
        f"({ONLINE_DOCS_ROOT}"
        f"/docs/compression_algorithms/Quantization.md#symmetric-quantization) for "
        f"more details.",
        enum=QUANTIZER_MODES,
    ),
    "bits": with_attributes(
        NUMBER,
        description="Bitwidth to quantize to. It is intended for manual bitwidth setting. Can be "
        "overridden by the `bits` parameter from the `precision` initializer section. "
        "An error occurs if it doesn't match a corresponding bitwidth constraints "
        "from the hardware configuration.",
        default=QUANTIZATION_BITS,
    ),
    "signed": with_attributes(
        BOOLEAN,
        description="Whether to use signed or unsigned input/output values for quantization. "
        "`true` will force the quantization to support signed values, `false` will "
        "force the quantization to only support input values with one and the same "
        "sign, and leaving this value unspecified (default) means relying "
        "on the initialization statistics to determine best approach. \n"
        "Note: If set to `false`, but the input "
        "values have differing signs during initialization, signed quantization "
        "will be performed instead.",
    ),
    "per_channel": with_attributes(
        BOOLEAN,
        description="Whether to quantize inputs of this quantizer "
        "per each channel of input tensor (per 0-th dimension for "
        "weight quantization, and per 1-st dimension for activation "
        "quantization).",
        default=QUANTIZATION_PER_CHANNEL,
    ),
}

UNIFIED_SCALE_OP_SCOPES_SPECIFIER_SCHEMA = {"type": "array", "items": ARRAY_OF_STRINGS}

RANGE_INIT_TYPES_VS_DESCRIPTIONS = {
    "mixed_min_max": "Minimum quantizer range initialized using minima of per-channel minima of "
    "the tensor to be quantized, maximum quantizer range initialized using "
    "maxima of per-channel maxima of the tensor to be quantized. Offline.",
    "min_max": "Minimum quantizer range initialized using global minimum of values in "
    "the tensor to be quantized, maximum quantizer range initialized using global maxima of the same"
    "values. Online.",
    "mean_min_max": "Minimum quantizer range initialized using averages (across every single initialization sample) "
    "of minima of values in "
    "the tensor to be quantized, maximum quantizer range initialized using maxima respectively. "
    "Offline.",
    "threesigma": "Quantizer minimum and maximum ranges set to be equal to +- 3 median absolute deviation from the "
    "median of the observed values in the tensor to be quantized. Offline.",
    "percentile": "Quantizer minimum and maximum ranges set to be equal to specified percentiles of the the observed "
    "values (across the entire initialization sample set) in the tensor to be quantized. Offline.",
    "mean_percentile": "Quantizer minimum and maximum ranges set to be equal to averaged (across every single "
    "initialization sample) specified percentiles of the the observed values in the tensor to be "
    "quantized. Offline.",
}

BASIC_RANGE_INIT_CONFIG_PROPERTIES = {
    "type": "object",
    "title": "global_range_init_configuration",
    "properties": {
        "num_init_samples": with_attributes(
            NUMBER,
            description="Number of samples from the training dataset to "
            "consume as sample model inputs for purposes of "
            "setting initial minimum and maximum quantization "
            "ranges.",
            default=NUM_INIT_SAMPLES,
        ),
        "type": with_attributes(
            annotated_enum(RANGE_INIT_TYPES_VS_DESCRIPTIONS),
            description="Type of the initializer - determines which "
            "statistics gathered during initialization will be "
            "used to initialize the quantization ranges.\n"
            "'Online' initializers do not have to "
            "store intermediate statistics in memory, while 'offline' do. Increasing the number "
            "of initialization samples for 'offline' initialization types will increase RAM "
            "overhead of applying NNCF to the model.\n"
            "Depending on whether the quantizer is configured to be per-tensor or per-channel, the "
            "statistics will be collected either on the basis of the set of the entire tensor values, or "
            "these will be collected and applied separately for each channel value subset.",
            default="mixed_min_max",
        ),
        "params": {
            "type": "object",
            "description": "Type-specific parameters of the initializer.",
            "properties": {
                "min_percentile": with_attributes(
                    NUMBER,
                    description="For 'percentile' and 'mean_percentile' types - specify the percentile "
                    "of input value histograms to be set as the initial "
                    "value for the quantizer input minimum.",
                    default=MIN_PERCENTILE,
                ),
                "max_percentile": with_attributes(
                    NUMBER,
                    description="For 'percentile' and 'mean_percentile' types - specify the percentile "
                    "of input value histograms to be set as the initial "
                    "value for the quantizer input maximum.",
                    default=MAX_PERCENTILE,
                ),
            },
        },
    },
    "additionalProperties": False,
}
PER_LAYER_RANGE_INIT_CONFIG_PROPERTIES = {
    "type": "object",
    "properties": {
        **BASIC_RANGE_INIT_CONFIG_PROPERTIES["properties"],  # type: ignore[dict-item]
        **SCOPING_PROPERTIES,
        "target_quantizer_group": with_attributes(
            STRING,
            description="The target group of quantizers for which "
            "the specified type of range initialization will "
            "be applied. If unspecified, then the range "
            "initialization of the given type will be applied to all quantizers.",
            enum=["activations", "weights"],
        ),
    },
}
RANGE_INIT_CONFIG_PROPERTIES = {
    "initializer": {
        "type": "object",
        "properties": {
            "range": {
                "description": "This initializer performs forward runs of the model to be quantized using "
                "samples from a user-supplied data loader to gather activation and weight "
                "tensor statistics within the network and use these to set up initial range parameters "
                "for quantizers.",
                "oneOf": [
                    BASIC_RANGE_INIT_CONFIG_PROPERTIES,
                    {
                        "type": "array",
                        "title": "per_layer_range_init_configuration",
                        "items": PER_LAYER_RANGE_INIT_CONFIG_PROPERTIES,
                    },
                ],
            },
        },
        "additionalProperties": False,
    },
}
BITWIDTH_ASSIGNMENT_MODE_SCHEMA = {
    "type": "string",
    "enum": ["strict", "liberal"],
    "default": "liberal",
    "description": "The mode for assignment bitwidth to activation quantizers. In the 'strict' mode,"
    "a group of quantizers that feed their output to one and more "
    "same modules as input (weight quantizers count as well) will have the same bitwidth in the "
    "'liberal' mode allows different precisions within the group.\n"
    "Bitwidth is assigned based on hardware constraints. If multiple "
    "variants are possible, the minimal compatible bitwidth is chosen.",
}

PRECISION_INIT_TYPES_VS_DESCRIPTION = {
    "hawq": f"Applies HAWQ algorithm to determine best bitwidths for each quantizer using a Hessian"
    f"calculation approach. For more details see "
    f"[Quantization.md]({ONLINE_DOCS_ROOT}/docs/compression_algorithms/Quantization.md#hawq)",
    "autoq": f"Applies AutoQ algorithm to determine best bitwidths for each quantizer using reinforcement learning. "
    f"For more details see "
    f"[Quantization.md]({ONLINE_DOCS_ROOT}/docs/compression_algorithms/Quantization.md#autoq)",
    "manual": "Allows to manually specify via following config options the exact bitwidth "
    "for each quantizer location. ",
}


PRECISION_INITIALIZER_SCHEMA = {
    "type": "object",
    "description": "This initializer performs advanced selection of bitwidth per each quantizer "
    "location, trying to achieve the best tradeoff between performance and "
    "quality of the resulting model.",
    "properties": {
        "required": ["type"],
        "type": with_attributes(
            annotated_enum(PRECISION_INIT_TYPES_VS_DESCRIPTION), description="Type of precision initialization."
        ),
        "bits": with_attributes(
            ARRAY_OF_NUMBERS,
            description="A list of bitwidth to choose from when performing precision "
            "initialization. Overrides bits constraints specified in "
            "`weight` and `activation` sections.",
            examples=[[4, 8]],
            default=PRECISION_INIT_BITWIDTHS,
        ),
        "num_data_points": with_attributes(
            NUMBER,
            description="Number of data points to iteratively estimate Hessian trace.",
            default=HAWQ_NUM_DATA_POINTS,
        ),
        "iter_number": with_attributes(
            NUMBER,
            description="Maximum number of iterations of Hutchinson algorithm to Estimate Hessian trace.",
            default=HAWQ_ITER_NUMBER,
        ),
        "tolerance": with_attributes(
            NUMBER,
            description="Minimum relative tolerance for stopping the Hutchinson "
            "algorithm. It's calculated  between mean average trace "
            "from the previous iteration and the current one.",
            default=HAWQ_TOLERANCE,
        ),
        "compression_ratio": with_attributes(
            NUMBER,
            description="For the `hawq` mode:\n"
            "The desired ratio between bit complexity of "
            "a fully INT8 model and a mixed-precision lower-bit "
            "one. On precision initialization stage the HAWQ "
            "algorithm chooses the most accurate "
            "mixed-precision configuration with a ratio no less "
            "than the specified. Bit complexity of the model "
            "is a sum of bit complexities for each quantized "
            "layer, which are a multiplication of FLOPS for "
            "the layer by the number of bits for its "
            "quantization.\n"
            "For the `autoq` mode:\n"
            "The target model size after quantization, relative to total "
            "parameters size in FP32. E.g. a uniform INT8-quantized model "
            "would have a `compression_ratio` equal to 0.25,"
            "and a uniform INT4-quantized model would have "
            "`compression_ratio` equal to 0.125.",
        ),
        "eval_subset_ratio": with_attributes(
            NUMBER,
            description="The desired ratio of dataloader to be iterated "
            "during each search iteration of AutoQ precision "
            "initialization. Specifically, this ratio applies "
            "to the registered autoq_eval_loader via "
            "register_default_init_args.",
            default=AUTOQ_EVAL_SUBSET_RATIO,
        ),
        "warmup_iter_number": with_attributes(
            NUMBER,
            description="The number of random policy at the beginning of "
            "of AutoQ precision initialization to populate "
            "replay buffer with experiences. This key is "
            "meant for internal testing use. Users need not "
            "to configure.",
            default=AUTOQ_WARMUP_ITER_NUMBER,
        ),
        "bitwidth_per_scope": {
            "type": "array",
            "items": {
                "type": "array",
                "items": [NUMBER, STRING],
                "description": "A tuple of a bitwidth and a scope of the quantizer to assign the bitwidth to.",
            },
            "description": "Manual settings for the quantizer bitwidths. Scopes are used to identify "
            "the quantizers.",
            "examples": [
                [
                    [2, "ResNet/NNCFConv2d[conv1]/conv2d_0|WEIGHT"],
                    [8, "ResNet/ReLU[relu]/relu__0|OUTPUT"],
                ]
            ],
        },
        "traces_per_layer_path": with_attributes(
            STRING,
            description="Path to serialized PyTorch Tensor with "
            "average Hessian traces per quantized modules. "
            "It can be used to accelerate mixed precision "
            "initialization by using average Hessian "
            "traces from previous run of HAWQ algorithm.",
        ),
        "dump_init_precision_data": with_attributes(
            BOOLEAN,
            description="Whether to dump data related to Precision "
            "Initialization algorithm. "
            "HAWQ dump includes bitwidth graph,"
            " average traces and different plots. "
            "AutoQ dump includes DDPG agent "
            "learning trajectory in tensorboard and "
            "mixed-precision environment metadata.",
            default=HAWQ_DUMP_INIT_PRECISION_DATA,
        ),
        "bitwidth_assignment_mode": BITWIDTH_ASSIGNMENT_MODE_SCHEMA,
    },
    "additionalProperties": False,
}

QUANTIZATION_INITIALIZER_SCHEMA = {
    "type": "object",
    "description": "Specifies the kind of pre-training initialization used for the quantization algorithm.\n"
    "Some kind of initialization is usually required so that the trainable quantization "
    "parameters have a better chance to get fine-tuned to values that result in good accuracy.",
    "properties": {
        "batchnorm_adaptation": BATCHNORM_ADAPTATION_SCHEMA,
        **RANGE_INIT_CONFIG_PROPERTIES["initializer"]["properties"],  # type: ignore[dict-item]
        "precision": PRECISION_INITIALIZER_SCHEMA,
    },
    "additionalProperties": False,
}
STAGED_QUANTIZATION_PARAMS = {
    "params": {
        "type": "object",
        "description": "Configures the staged quantization compression scheduler for the quantization algorithm. "
        "The quantizers will not be applied until a given epoch count is reached.",
        "properties": {
            "batch_multiplier": with_attributes(
                NUMBER,
                description="Gradients will be accumulated for this number of "
                "batches before doing a 'backward' call. Increasing "
                "this may improve training quality, since binarized "
                "networks exhibit noisy gradients and their training "
                "requires larger batch sizes than could be accommodated "
                "by GPUs.",
            ),
            "activations_quant_start_epoch": with_attributes(
                NUMBER,
                description="A zero-based index of the epoch, upon reaching which "
                "the activations will start to be quantized.",
                default=ACTIVATIONS_QUANT_START_EPOCH,
            ),
            "weights_quant_start_epoch": with_attributes(
                NUMBER,
                description="Epoch index upon which the weights will start to be quantized.",
                default=WEIGHTS_QUANT_START_EPOCH,
            ),
            "lr_poly_drop_start_epoch": with_attributes(
                NUMBER,
                description="Epoch index upon which the learning rate will "
                "start to be dropped. If unspecified, "
                "learning rate will not be dropped.",
            ),
            "lr_poly_drop_duration_epochs": with_attributes(
                NUMBER,
                description="Duration, in epochs, of the learning rate dropping process.",
                default=LR_POLY_DURATION_EPOCHS,
            ),
            "disable_wd_start_epoch": with_attributes(
                NUMBER,
                description="Epoch to disable weight decay in the optimizer. "
                "If unspecified, weight decay will not be disabled.",
            ),
            "base_lr": with_attributes(
                NUMBER, description="Initial value of learning rate.", default=STAGED_QUANTIZATION_BASE_LR
            ),
            "base_wd": with_attributes(
                NUMBER, description="Initial value of weight decay.", default=STAGED_QUANTIZATION_BASE_WD
            ),
        },
        "additionalProperties": False,
    }
}

QUANTIZATION_PRESETS_SCHEMA = {"type": "string", "enum": ["performance", "mixed"]}

QUANTIZER_GROUP_PROPERTIES = {
    **QUANTIZER_CONFIG_PROPERTIES,
    **SCOPING_PROPERTIES,
    "logarithm_scale": with_attributes(
        BOOLEAN,
        description="Whether to use log of scale as the optimization parameter "
        "instead of the scale itself. This serves as an optional "
        "regularization opportunity for training quantizer scales.",
        default=QUANTIZATION_LOGARITHM_SCALE,
    ),
}

WEIGHTS_GROUP_SCHEMA = {
    "type": "object",
    "properties": {
        **QUANTIZER_GROUP_PROPERTIES,
    },
    "additionalProperties": False,
}
ACTIVATIONS_GROUP_SCHEMA = {
    "type": "object",
    "properties": {
        **QUANTIZER_GROUP_PROPERTIES,
        "unified_scale_ops": with_attributes(
            UNIFIED_SCALE_OP_SCOPES_SPECIFIER_SCHEMA,
            description="Specifies operations in the model which will share "
            "the same quantizer module for activations. This "
            "is helpful in case one and the same quantizer scale "
            "is required for each input of this operation. Each "
            "sub-array will define a group of model operation "
            "inputs that have to share a single actual "
            "quantization module, each entry in this subarray "
            "should correspond to exactly one node in the NNCF "
            "graph and the groups should not overlap. The final "
            "quantizer for each sub-array will be associated "
            "with the first element of this sub-array.",
        ),
    },
    "additionalProperties": False,
}

OVERFLOW_FIX_OPTIONS = ["enable", "disable", "first_layer_only"]

QUANTIZATION_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "description": f"Applies quantization on top of the input model, simulating future low-precision execution "
    f"specifics, and selects the quantization layout and parameters to strive for the best possible "
    f"quantized model accuracy and performance. "
    f"\nSee [Quantization.md]"
    f"({ONLINE_DOCS_ROOT}"
    f"/docs/compression_algorithms/Quantization.md) and the rest of this schema for "
    f"more details and parameters.",
    "properties": {
        "algorithm": {"const": QUANTIZATION_ALGO_NAME_IN_CONFIG},
        "initializer": QUANTIZATION_INITIALIZER_SCHEMA,
        "preset": with_attributes(
            QUANTIZATION_PRESETS_SCHEMA,
            description="The preset defines the quantization schema for weights and activations. "
            "The 'performance' mode sets up symmetric weight and activation "
            "quantizers. The 'mixed' mode utilizes symmetric weight quantization and "
            "asymmetric activation quantization.",
            default=QUANTIZATION_PRESET,
        ),
        "quantize_inputs": with_attributes(
            BOOLEAN,
            description="Whether the model inputs should be immediately quantized prior "
            "to any other model operations.",
            default=QUANTIZE_INPUTS,
        ),
        "quantize_outputs": with_attributes(
            BOOLEAN, description="Whether the model outputs should be additionally quantized.", default=QUANTIZE_OUTPUTS
        ),
        "weights": with_attributes(
            WEIGHTS_GROUP_SCHEMA, description="Constraints to be applied to model weights quantization only."
        ),
        "activations": with_attributes(
            ACTIVATIONS_GROUP_SCHEMA, description="Constraints to be applied to model activations quantization only."
        ),
        "scope_overrides": {
            "type": "object",
            "description": "This option is used to specify overriding quantization constraints for specific scope,"
            "e.g. in case you need to quantize a single operation differently than the rest of the "
            "model. Any other automatic or group-wise settings will be overridden.",
            "examples": [
                {
                    "weights": {
                        "QuantizeOutputsTestModel/NNCFConv2d[conv5]/conv2d_0": {
                            "mode": "asymmetric",
                        },
                        "activations": {
                            "{re}.*conv_first.*": {"mode": "asymmetric"},
                            "{re}.*conv_second.*": {"mode": "symmetric"},
                        },
                    }
                }
            ],
            "properties": {
                "weights": {
                    "type": "object",
                    "patternProperties": {
                        ".*": {
                            "type": "object",
                            "properties": {
                                **QUANTIZER_CONFIG_PROPERTIES,
                                **RANGE_INIT_CONFIG_PROPERTIES,
                            },
                            "additionalProperties": False,
                        },
                    },
                },
                "activations": {
                    "type": "object",
                    "patternProperties": {
                        ".*": {
                            "type": "object",
                            "properties": {
                                **QUANTIZER_CONFIG_PROPERTIES,
                                **RANGE_INIT_CONFIG_PROPERTIES,
                            },
                            "additionalProperties": False,
                        },
                    },
                },
            },
            "additionalProperties": False,
        },
        "export_to_onnx_standard_ops": with_attributes(
            BOOLEAN,
            description="[Deprecated] Determines how should the additional quantization "
            "operations be exported into the ONNX format. Set "
            "this to true to export to ONNX "
            "standard QuantizeLinear-DequantizeLinear "
            "node pairs (8-bit quantization only) or to false "
            "to export to OpenVINO-supported FakeQuantize ONNX"
            "(all quantization settings supported).",
            default=QUANTIZATION_EXPORT_TO_ONNX_STANDARD_OPS,
        ),
        "overflow_fix": with_attributes(
            STRING,
            description="This option controls whether to apply the overflow "
            "issue fix for the appropriate NNCF config or not. "
            "If set to `disable`, the fix will not be applied. "
            "If set to `enable` or `first_layer_only`, "
            "while appropriate target_devices are chosen, "
            "the fix will be applied to all layers or to the first "
            "convolutional layer respectively.",
            enum=OVERFLOW_FIX_OPTIONS,
            default=QUANTIZATION_OVERFLOW_FIX,
        ),
        **STAGED_QUANTIZATION_PARAMS,
        **SCOPING_PROPERTIES,
        **COMPRESSION_LR_MULTIPLIER_PROPERTY,
    },
    "additionalProperties": False,
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum
from enum import auto


class TensorBackend(Enum):
    """
    Enum representing the different tensor backends.
    """

    numpy = auto()
    torch = auto()
    ov = auto()


class TensorDataType(Enum):
    """
    Enum representing the different tensor data types.
    """

    float16 = auto()
    bfloat16 = auto()
    float32 = auto()
    float64 = auto()
    int8 = auto()
    int32 = auto()
    int64 = auto()
    uint8 = auto()
    uint4 = auto()
    int4 = auto()

    def is_float(self):
        """
        :return: True if the tensor data type is a floating-point type, else False.
        """
        return self in [TensorDataType.float16, TensorDataType.bfloat16, TensorDataType.float32, TensorDataType.float64]


class TensorDeviceType(Enum):
    """
    Enum representing the different tensor device types.
    """

    CPU = auto()
    GPU = auto()


@dataclass
class TypeInfo:
    """
    The class represents the numerical properties of a floating point types.

    :param eps: The smallest representable number such that 1.0 + eps != 1.0.
    :param max: The largest representable number.
    :param min: The smallest representable number (typically -max).
    """

    eps: float
    max: float
    min: float



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import Logger
from typing import Generic, Iterable, Iterator, Optional, TypeVar

from nncf.common.logging import nncf_logger

TObj = TypeVar("TObj")


class ProgressBar(Generic[TObj]):
    """
    A basic progress bar specifically for the logging.
    It does not print at the same line, instead it logs multiple lines intentionally to avoid multiprocessing issues.
    Works as a decorator for iterable.
    :param iterable: iterable to decorate with a progressbar.
    :param logger: logging to print a progress, nncf_logger by default
    :param desc: prefix for the printed line, empty by default
    :param num_lines:  defines the number of lines to log
    :param total: the expected total number of iterations
    """

    def __init__(
        self,
        iterable: Iterable[TObj],
        logger: Logger = nncf_logger,
        desc: str = "",
        num_lines: int = 10,
        total: Optional[int] = None,
    ):
        self._logger = logger
        self._iterable = iterable
        self._desc = desc
        self._num_lines = num_lines

        self._index: int = 0
        self._width = 16
        self._is_enabled = False
        self._total = None
        if total is not None:
            if not isinstance(total, (int, float)) or total <= 0:
                logger.error(
                    "Progress bar is disabled because the expected total number of iterations is invalid: "
                    "it should be an integer and more than 0"
                )
                return
            self._total = int(total)

        if iterable is not None and self._total is None:
            try:
                self._total = len(iterable)  # type: ignore[arg-type]
            except (TypeError, AttributeError):
                logger.error(
                    "Progress bar is disabled because the given iterable is invalid: "
                    "it does not implement __len__ method"
                )
                return

        if not isinstance(num_lines, int) or num_lines <= 1:
            logger.error(
                "Progress bar is disabled because the given number of lines for logging is invalid: "
                "it should be an integer and more than 1"
            )
            return

        self._step = max(1, self._total // (self._num_lines - 1))
        self._is_enabled = True

    def __iter__(self) -> Iterator[TObj]:
        for obj in self._iterable:
            yield obj
            if self._is_enabled:
                self._print_next()

    def _print_next(self) -> None:
        self._index += 1
        if self._total is None or self._index > self._total:
            return

        if self._index % self._step == 0 or self._index == self._total:
            num_filled = int(self._index * self._width / self._total)
            num_empty = self._width - num_filled
            filled = "█" * num_filled
            empty = " " * num_empty
            self._logger.info(
                "{desc} |{filled}{empty}| {index} / {total}".format(
                    desc=self._desc, filled=filled, empty=empty, index=self._index, total=self._total
                )
            )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict


class Registry:
    REGISTERED_NAME_ATTR = "_registered_name"

    def __init__(self, name: str, add_name_as_attr: bool = False):
        self._name = name
        self._registry_dict: Dict[str, Any] = {}
        self._add_name_as_attr = add_name_as_attr

    @property
    def registry_dict(self) -> Dict[str, Any]:
        return self._registry_dict

    def values(self) -> Any:
        return self._registry_dict.values()

    def _register(self, obj: Any, name: str) -> None:
        if name in self._registry_dict:
            raise KeyError("{} is already registered in {}".format(name, self._name))
        self._registry_dict[name] = obj

    def register(self, name: str = None) -> Callable[[Any], Any]:
        def wrap(obj: Any) -> Any:
            cls_name = name
            if cls_name is None:
                cls_name = obj.__name__
            if self._add_name_as_attr:
                setattr(obj, self.REGISTERED_NAME_ATTR, name)
            self._register(obj, cls_name)
            return obj

        return wrap

    def get(self, name: str) -> Any:
        if name not in self._registry_dict:
            self._key_not_found(name)
        return self._registry_dict[name]

    def _key_not_found(self, name: str) -> None:
        raise KeyError("{} is unknown type of {} ".format(name, self._name))

    def __contains__(self, item: Any) -> bool:
        return item in self._registry_dict.values()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from enum import Enum
from typing import Any, Callable, TypeVar

from packaging import version

import nncf

TModel = TypeVar("TModel")

try:
    import openvino  # type: ignore # noqa: F401

    _OPENVINO_AVAILABLE = True
except ImportError:
    _OPENVINO_AVAILABLE = False


class BackendType(Enum):
    TORCH = "Torch"
    TORCH_FX = "TorchFX"
    TENSORFLOW = "Tensorflow"
    ONNX = "ONNX"
    OPENVINO = "OpenVINO"


def result_verifier(func: Callable[[TModel], bool]) -> Callable[..., None]:
    def verify_result(*args: Any, **kwargs: Any):  # type: ignore
        try:
            return func(*args, **kwargs)
        except Exception:
            return False

    return verify_result


@result_verifier
def is_torch_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of torch.nn.Module and not a torch.fx.GraphModule, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of torch.nn.Module and not torch.fx.GraphModule, otherwise False.
    """
    import torch
    import torch.fx

    return not isinstance(model, torch.fx.GraphModule) and isinstance(model, torch.nn.Module)


@result_verifier
def is_torch_fx_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of torch.fx.GraphModule, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of torch.fx.GraphModule, otherwise False.
    """
    import torch.fx

    return isinstance(model, torch.fx.GraphModule)


@result_verifier
def is_tensorflow_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of tensorflow.Module, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of tensorflow.Module, otherwise False.
    """
    import tensorflow  # type: ignore

    return isinstance(model, tensorflow.Module)


@result_verifier
def is_onnx_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of onnx.ModelProto, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of onnx.ModelProto, otherwise False.
    """
    import onnx  # type: ignore

    return isinstance(model, onnx.ModelProto)


@result_verifier
def is_openvino_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of openvino.runtime.Model, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of openvino.runtime.Model, otherwise False.
    """
    import openvino.runtime as ov  # type: ignore

    return isinstance(model, ov.Model)


@result_verifier
def is_openvino_compiled_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of openvino.runtime.CompiledModel, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of openvino.runtime.CompiledModel, otherwise False.
    """
    import openvino.runtime as ov

    return isinstance(model, ov.CompiledModel)


def get_backend(model: TModel) -> BackendType:
    """
    Returns the NNCF backend name string inferred from the type of the model object passed into this function.

    :param model: The framework-specific model.
    :return: A BackendType representing the correct NNCF backend to be used when working with the framework.
    """

    verify_map = {
        is_torch_fx_model: BackendType.TORCH_FX,
        is_torch_model: BackendType.TORCH,
        is_tensorflow_model: BackendType.TENSORFLOW,
        is_onnx_model: BackendType.ONNX,
        is_openvino_model: BackendType.OPENVINO,
    }

    for backend_call, backend in verify_map.items():
        if backend_call(model):
            return backend

    raise nncf.UnsupportedBackendError(
        "Could not infer the backend framework from the model type because "
        "the framework is not available or corrupted, or the model type is unsupported. "
    )


def copy_model(model: TModel) -> TModel:
    """
    Function to create copy of the backend-specific model.

    :param model: the backend-specific model instance
    :return: Copy of the backend-specific model instance.
    """
    model_backend = get_backend(model)
    if model_backend == BackendType.OPENVINO:
        # TODO(l-bat): Remove after fixing ticket: 100919
        return model.clone()  # type: ignore
    if model_backend == BackendType.TENSORFLOW:
        # deepcopy and tensorflow.keras.models.clone_model does not work correctly on 2.8.4 version
        from nncf.tensorflow.graph.model_transformer import TFModelTransformer
        from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout

        model = TFModelTransformer(model).transform(TFTransformationLayout())
        return model
    return deepcopy(model)


def is_openvino_available() -> bool:
    """
    Check if OpenVINO is available.
    :return: True if openvino package is installed, False otherwise.
    """
    return _OPENVINO_AVAILABLE


def is_openvino_at_least(version_str: str) -> bool:
    """
    Check if OpenVINO version is at least the specified one.

    :param version_str: The version string to compare with the installed OpenVINO version. For example "2025.1".
    :return: True if the installed OpenVINO version is at least the specified one, False otherwise.
    """
    if not _OPENVINO_AVAILABLE:
        return False

    openvino_version = version.parse(openvino.__version__.split("-")[0])
    return version.parse(version_str) <= openvino_version



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.commands import TransformationType
from nncf.experimental.tensorflow.graph.transformations.layout import TFTransformationLayoutV2
from nncf.experimental.tensorflow.nncf_network import NNCFNetwork


class TFModelTransformerV2(ModelTransformer):
    """
    Applies transformations to the NNCF network.

    The `TFModelTransformerV2` does not modify the model config to insert
    compression operations to the model. The `TFModelTransformerV2` adds
    pre-hook or post-hook to the wrapped TF operation instead of that.
    In this way, compression operation falls into the model.
    """

    def __init__(self, model: NNCFNetwork):
        """
        Initializes the model transformer.

        :param model: NNCF network.
        """
        super().__init__(model)

    def transform(self, transformation_layout: TFTransformationLayoutV2) -> NNCFNetwork:
        """
        Applies transformations to the model.

        :param transformation_layout: An instance of `TransformationLayout` that
            includes a list of transformations to be applied to the NNCF network.
        :return: The transformed NNCF network.
        """
        for command in transformation_layout.transformations:
            if command.type == TransformationType.INSERT:
                self._model.insert_at_point(command.target_point, command.insertion_objects)
            elif command.type == TransformationType.REMOVE:
                # TODO(andrey-churkin): Add support
                pass
            else:
                raise ValueError(f"Transformation type {command.type} does not support.")

        return self._model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple, Union

import numpy as np
import openvino as ov

from nncf.tensor import Tensor
from nncf.tensor import TensorDataType
from nncf.tensor.definitions import TensorBackend
from nncf.tensor.definitions import TensorDeviceType
from nncf.tensor.functions import numeric

DTYPE_MAP = {
    TensorDataType.float16: ov.Type.f16,
    TensorDataType.bfloat16: ov.Type.bf16,
    TensorDataType.float32: ov.Type.f32,
    TensorDataType.float64: ov.Type.f64,
    TensorDataType.int8: ov.Type.i8,
    TensorDataType.int32: ov.Type.i32,
    TensorDataType.int64: ov.Type.i64,
    TensorDataType.uint8: ov.Type.u8,
    TensorDataType.uint4: ov.Type.u4,
    TensorDataType.int4: ov.Type.i4,
}

DTYPE_MAP_REV = {v: k for k, v in DTYPE_MAP.items()}


@numeric.device.register(ov.Tensor)
def _(a: ov.Tensor) -> TensorDeviceType:
    return TensorDeviceType.CPU


@numeric.backend.register(ov.Tensor)
def _(a: ov.Tensor) -> TensorBackend:
    return TensorBackend.ov


@numeric.astype.register(ov.Tensor)
def _(a: ov.Tensor, dtype: TensorDataType) -> ov.Tensor:
    if a.get_element_type() in [ov.Type.bf16, ov.Type.i4, ov.Type.u4] or dtype in [
        TensorDataType.bfloat16,
        TensorDataType.int4,
        TensorDataType.uint4,
    ]:
        # Cannot cast to/from bfloat16, uint4, int4 directly
        return _astype_ov(a, dtype)
    return ov.Tensor(numeric.astype(a.data, dtype))


@numeric.dtype.register(ov.Tensor)
def _(a: ov.Tensor) -> TensorDataType:
    return DTYPE_MAP_REV[a.get_element_type()]


@numeric.size.register(ov.Tensor)
def _(a: ov.Tensor) -> int:
    return a.size


@numeric.reshape.register(ov.Tensor)
def _(a: ov.Tensor, shape: Union[int, Tuple[int, ...]]) -> ov.Tensor:
    return ov.Tensor(a.data.reshape(shape), shape, a.get_element_type())


@numeric.as_numpy_tensor.register(ov.Tensor)
def _(a: ov.Tensor) -> np.ndarray:
    # Cannot convert bfloat16, uint4, int4 to numpy directly
    a_dtype = DTYPE_MAP_REV[a.get_element_type()]
    if a_dtype in [TensorDataType.bfloat16, TensorDataType.uint4, TensorDataType.int4]:
        dtype = TensorDataType.float32
        if a_dtype == TensorDataType.uint4:
            dtype = TensorDataType.uint8
        elif a_dtype == TensorDataType.int4:
            dtype = TensorDataType.int8
        a = _astype_ov(a, dtype)
    return a.data


def _astype_ov(a: ov.Tensor, dtype: TensorDataType) -> ov.Tensor:
    """
    Cast to a different data type using an OpenVINO model.
    :param a: Tensor to cast.
    :param dtype: Data type to cast to.
    :return: Casted openvino tensor.
    """

    from nncf.openvino.optimized_functions import astype

    return astype(Tensor(a), dtype).data



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.fx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.model_utils import get_target_point
from nncf.experimental.torch.fx.node_utils import get_bias_value
from nncf.experimental.torch.fx.node_utils import is_node_with_bias
from nncf.experimental.torch.fx.transformations import constant_update_transformation_builder
from nncf.quantization.algorithms.fast_bias_correction.backend import FastBiasCorrectionAlgoBackend
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.commands import PTModelExtractionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import is_quantized_weights
from nncf.torch.tensor_statistics.collectors import get_mean_statistic_collector


class FXFastBiasCorrectionAlgoBackend(FastBiasCorrectionAlgoBackend):
    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        return get_target_point(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> FXApplyTransformationCommand:
        return FXApplyTransformationCommand(
            constant_update_transformation_builder(node, bias_value.data, input_port_id=2)
        )

    @staticmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> PTModelExtractionCommand:
        return PTModelExtractionCommand([input_ids[0][0]], [output_ids[0][0]])

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size)

    @staticmethod
    def get_sub_input_output_names(subgraph: torch.fx.GraphModule) -> Tuple[Optional[int], int]:
        # Pytorch does not have name for extracted node
        return None, 0

    @staticmethod
    def create_input_data(shape: Tuple[int], data: List[Tensor], input_name: str, channel_axis: int) -> torch.Tensor:
        blob = torch.zeros(shape, dtype=data[0].data.dtype, device=data[0].data.device)
        for j, idx in enumerate(np.ndindex(blob.shape[channel_axis])):
            index = tuple(slice(None) if i != channel_axis else idx for i in range(blob.ndim))
            blob[index] = data[j].data
        return blob

    @staticmethod
    def get_bias_value(node: NNCFNode, nncf_graph: NNCFGraph, model: torch.fx.GraphModule) -> Tensor:
        return get_bias_value(node, nncf_graph, model)

    @staticmethod
    def get_activation_port_ids_for_bias_node(node: NNCFNode) -> Tuple[int, int]:
        return 0, 0

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: int) -> Tensor:
        return Tensor(raw_data[output_name])

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_quantized_weights(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def get_node_names_for_input_output_statistics(node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[str, str]:
        return node.node_name, node.node_name

    @staticmethod
    def get_activation_channel_axis(node: NNCFNode, pord_id: int, input_shape: Tuple[int]) -> int:
        return node.metatype.output_channel_axis



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, TypeVar

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.data.dataset import Dataset
from nncf.errors import UnsupportedBackendError
from nncf.quantization.algorithms.layerwise.iterator import LayerwiseIterator
from nncf.quantization.algorithms.layerwise.scheduler import LayerwiseScheduler
from nncf.quantization.algorithms.layerwise.scheduler import LayerwiseStep
from nncf.quantization.algorithms.layerwise.scheduler import NodeOutputPort
from nncf.tensor import Tensor

TModel = TypeVar("TModel")


class LayerwiseEngine:
    """
    A class to perform layer-wise processing of a model using a specified strategy.

    This class allows for layer-wise analysis and manipulation of a model. It supports
    iteration through target nodes and collecting inputs or outputs of target nodes.
    """

    def __init__(
        self,
        subset_size: int = 100,
        collect_inputs: bool = True,
    ):
        """
        :param subset_size: Number of samples from the dataset to use, defaults to 100.
        :param collect_inputs: Whether to collect inputs for the layers, defaults to True.

        """
        self._scheduler = LayerwiseScheduler()
        self._subset_size = subset_size
        self._collect_inputs = collect_inputs

        self._backend_entity = None
        self._algorithm_key = f"LayerwiseEngine_{hash(self)}"

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic for the engine.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.layerwise.openvino_backend import OVLayerwiseEngineBackend

            self._backend_entity = OVLayerwiseEngineBackend()
        else:
            raise UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def _get_statistics(self, statistic_points: StatisticPointsContainer, node_name: str, port_id: int) -> List[Tensor]:
        """
        Collects statistic values for the given node and port id.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :param port_id: Port id for statistics collection.
        :return: Collected list of tensor data.
        """

        def input_filter_func(point):
            # For the floating-point statistics collected in POST_LAYER style,
            # we also need to determine the output port id.
            # For the cases when the layer has more than one (0) output port.
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
                and point.target_point.port_id == port_id
            )

        res = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, input_filter_func, self._algorithm_key
        ):
            for value in tensor_collector.get_statistics().values:
                res.append(Tensor(value))
        return res

    def _create_cache(
        self, outputs: List[NodeOutputPort], statistic_points: StatisticPointsContainer
    ) -> Dict[NodeOutputPort, List[Tensor]]:
        """
        Creates a cache for the outputs using the provided statistic points.

        :param outputs: List of node output ports.
        :param statistic_points: Container holding statistic points.
        :return: Dictionary mapping node output ports to lists of tensor data.
        """
        cache = {}
        for out in outputs:
            x = self._get_statistics(statistic_points, out.node_name, out.output_port)
            if not x:
                raise RuntimeError(f"Statistics for {out.node_name} node on {out.output_port} port is not provided.")
            cache[out] = x
        return cache

    def _get_outputs_following_model_inputs(
        self, graph: NNCFGraph, schedule: List[LayerwiseStep]
    ) -> List[NodeOutputPort]:
        """
        Identifies the outputs that follow the model inputs.

        :param graph: The model graph.
        :param schedule: List of layer-wise steps defining the schedule.
        :return: List of node output ports following model inputs.
        """
        input_node_names = [node.node_name for node in graph.get_input_nodes()]
        outputs = set()
        for step in schedule:
            graph_inputs = set()
            for sub_input in step.subgraph_inputs:
                if sub_input.node_name in input_node_names:
                    graph_inputs.add(sub_input)
            num_graph_inputs = len(graph_inputs)
            if num_graph_inputs == len(step.subgraph_inputs):
                outputs |= set(step.subgraph_outputs)
            else:
                for sub_input in graph_inputs:
                    outputs.add(sub_input)
        return list(outputs)

    def create_iterator_through_target_nodes(
        self,
        model: TModel,
        graph: NNCFGraph,
        target_nodes: List[NNCFNode],
        dataset: Dataset,
        statistic_points: Optional[StatisticPointsContainer] = None,
    ) -> LayerwiseIterator:
        """
        Create the iterator through the target nodes of the model.

        :param model: The model to be iterated over.
        :param graph: The model graph.
        :param target_nodes: List of target nodes to iterate through.
        :param dataset: The dataset to be used for obtaining inputs to the model.
        :param statistic_points: Optional container holding statistic points.
        :return: An iterator yielding inputs or outputs for each target node of the model.
        """
        self._set_backend_entity(model)

        schedule = self._scheduler.schedule(graph, target_nodes, self._collect_inputs)
        cache = None
        if statistic_points is not None:
            outputs = self._get_outputs_following_model_inputs(graph, schedule)
            cache = self._create_cache(outputs, statistic_points)
        return self._backend_entity.create_layerwise_iterator(model, graph, schedule, dataset, self._subset_size, cache)

    def get_statistic_points(
        self, model: TModel, graph: NNCFGraph, target_nodes: List[NNCFNode]
    ) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: The model for statistics collection.
        :param graph: The model graph.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """
        self._set_backend_entity(model)

        schedule = self._scheduler.schedule(graph, target_nodes, self._collect_inputs)
        outputs = self._get_outputs_following_model_inputs(graph, schedule)

        statistic_container = StatisticPointsContainer()
        for node_name, output_port_id in outputs:
            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, node_name, port_id=output_port_id
            )
            stat_collector = self._backend_entity.raw_statistic_collector(num_samples=self._subset_size)
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )
        return statistic_container



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from itertools import chain
from typing import Any, Dict, Iterable, List, Optional, Union

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.modules.batchnorm import _BatchNorm

import nncf
from nncf import nncf_logger
from nncf.common.graph.graph import NNCFNode
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.model_graph_manager import get_const_data
from nncf.torch.model_graph_manager import get_const_node
from nncf.torch.model_graph_manager import get_fake_quantizer
from nncf.torch.nncf_network import NNCFNetwork

BATCH_NORM_CLASSES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


CONV_METATYPES = (
    om.PTConv1dMetatype,
    om.PTConv2dMetatype,
    om.PTConv3dMetatype,
    om.PTDepthwiseConv1dSubtype,
    om.PTDepthwiseConv2dSubtype,
    om.PTDepthwiseConv3dSubtype,
)

CONV_TRANSPOSE_METATYPES = (
    om.PTConvTranspose1dMetatype,
    om.PTConvTranspose2dMetatype,
    om.PTConvTranspose3dMetatype,
)


class ExtractedFunc(nn.Module):
    def __init__(
        self,
        fn_name: str,
        kwargs: Dict[str, Any],
    ) -> None:
        super().__init__()
        self.fn_name = fn_name
        self.kwargs = kwargs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return getattr(F, self.fn_name)(input=x, **self.kwargs)


def extract_conv(
    input_node: NNCFNode,
    output_node: NNCFNode,
    model: NNCFNetwork,
) -> ExtractedFunc:
    """
    Extracts a convolutional layer from an NNCF graph and constructs an ExtractedFunc module.

    :param input_nodes: The name of input node.
    :param output_nodes: The name of output node.
    :param model: The NNCF network containing the layer.
    :return: The extracted convolutional layer as an ExtractedFunc module.
    """
    graph = model.nncf.get_graph()
    weight_node = get_const_node(input_node, input_node.metatype.weight_port_ids[0], graph)
    weight = get_const_data(weight_node, model)
    w_fq = get_fake_quantizer(input_node, input_node.metatype.weight_port_ids[0], model)
    bias_node = get_const_node(input_node, input_node.metatype.bias_port_id, graph)
    bias = get_const_data(bias_node, model) if bias_node is not None else None

    with torch.no_grad():
        e_weight = w_fq(weight) if w_fq else weight

    if input_node.metatype in CONV_METATYPES:
        kwargs = {
            "weight": e_weight.clone(),
            "bias": bias.clone() if bias is not None else bias,
            "stride": input_node.layer_attributes.stride,
            "padding": input_node.layer_attributes.padding_values,
            "dilation": input_node.layer_attributes.dilations,
            "groups": input_node.layer_attributes.groups,
        }
    elif input_node.metatype in CONV_TRANSPOSE_METATYPES:
        kwargs = {
            "weight": e_weight.clone(),
            "bias": bias.clone() if bias is not None else bias,
            "stride": input_node.layer_attributes.stride,
            "padding": input_node.layer_attributes.padding_values,
            "output_padding": input_node.layer_attributes.output_padding_values,
            "dilation": input_node.layer_attributes.dilations,
        }

    extracted_module = ExtractedFunc(input_node.node_type, kwargs)

    if input_node != output_node:
        extracted_module = try_to_fuse_conv(input_node, output_node, model, extracted_module)

    return extracted_module


def _find_parent_class(cls: type, parent_classes: Iterable[type]) -> Optional[type]:
    """
    Finds the first parent class of the given class that is present in the list of possible parent classes.

    :param cls: The class whose parent to find.
    :param parent_classes: A list of potential parent classes.
    :return: The first matching parent class, or None if no match is found.
    """
    for exp_cls in parent_classes:
        if issubclass(cls, exp_cls):
            return exp_cls
    return None


def extract_bn(node: NNCFNode, model: NNCFNetwork) -> Optional[Union[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d]]:
    """
    Extract batch_norm operation.
    If source modules inhered from nn.BatchNorm1d, nn.BatchNorm2d, or nn.BatchNorm3d return torch BatchNorm module.

    :param node: Target batch_norm node.
    :param model: Source model.
    :return: BatchNorm module with same attributes and parameters from source module or None.
    """
    bn_module: _BatchNorm = model.nncf.get_containing_module(node.node_name)
    bn_class = _find_parent_class(bn_module.__class__, BATCH_NORM_CLASSES)
    if bn_class is None:
        nncf_logger.debug(f"Module associated with {node} should be inhered from one of {BATCH_NORM_CLASSES}")
        return None

    extracted_bn: _BatchNorm = bn_class(
        num_features=bn_module.num_features,
        eps=bn_module.eps,
        momentum=bn_module.momentum,
        affine=bn_module.affine,
        track_running_stats=bn_module.track_running_stats,
        device=bn_module.weight.device,
        dtype=bn_module.weight.dtype,
    )

    # Copy named parameters and buffer that exists in native BatchNorm module from module in the module.
    for name, _ in chain(extracted_bn.named_parameters(), extracted_bn.named_buffers()):
        setattr(extracted_bn, name, deepcopy(getattr(bn_module, name)))
    extracted_bn.eval()
    return extracted_bn


def try_to_fuse_conv(
    input_node: NNCFNode, output_node: NNCFNode, model: NNCFNetwork, extracted_module: nn.Module
) -> nn.Module:
    """
    Fused convolution operation with the next batch norm node if possible,

    :param input_node: Input subgraph node.
    :param output_node: Output subgraph node (fused with input node).
    :param model: Source model.
    :param extracted_module: Extracted module.
    """
    next_nodes = model.nncf.get_graph().get_next_nodes(input_node)

    if len(next_nodes) != 1:
        return extracted_module

    if output_node != next_nodes[0]:
        raise nncf.InternalError(f"Output node {output_node} not found after {input_node}")

    if next_nodes[0].metatype != om.PTBatchNormMetatype:
        raise nncf.InternalError("Supported only BatchNorm layers")

    extracted_bn = extract_bn(next_nodes[0], model)
    if extracted_bn is None:
        nncf_logger.debug(
            f"Can`t extract fused batchnorm module for {input_node.node_name},"
            " module that contain batchnorm operator should be inhered from one of {BATCH_NORM_CLASSES}."
        )
        return None
    return nn.Sequential(extracted_module, extracted_bn)


def extract_model(model: NNCFNetwork, input_nodes: List[str], output_nodes: List[str]) -> Optional[nn.Module]:
    """
    Extracts a submodule from a given NNCF network containing only the nodes from the input to the output node.

    :param model: The NNCF network to extract the submodule from.
    :param input_nodes: List containing names of the input nodes for the submodule.
    :param output_nodes: List containing names of the output nodes for the submodule.
    :return: An nn.Module containing the extracted submodel, or None if extraction is not supported.
    """

    if len(input_nodes) != 1 or len(output_nodes) != 1:
        raise nncf.InternalError("input_nodes and output_nodes should contain only one node.")

    graph = model.nncf.get_graph()
    input_node = graph.get_node_by_name(input_nodes[0])
    output_node = graph.get_node_by_name(output_nodes[0])

    if input_node.metatype in CONV_METATYPES + CONV_TRANSPOSE_METATYPES:
        return extract_conv(input_node, output_node, model)

    nncf_logger.debug(f"Can`t extract module for {input_node.node_name}")
    return None



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Type

from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.tensorflow.graph.metatypes import keras_layers as layer_metatypes
from nncf.tensorflow.graph.metatypes import tf_ops as op_metatypes

ALL_LAYER_METATYPES_WITH_WEIGHTS = [
    layer_metatypes.TFConv1DLayerMetatype,
    layer_metatypes.TFConv2DLayerMetatype,
    layer_metatypes.TFConv3DLayerMetatype,
    layer_metatypes.TFDepthwiseConv1DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv3DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DLayerMetatype,
    layer_metatypes.TFConv1DTransposeLayerMetatype,
    layer_metatypes.TFConv2DTransposeLayerMetatype,
    layer_metatypes.TFConv3DTransposeLayerMetatype,
    layer_metatypes.TFDenseLayerMetatype,
    layer_metatypes.TFBatchNormalizationLayerMetatype,
    layer_metatypes.TFSeparableConv1DLayerMetatype,
    layer_metatypes.TFSeparableConv2DLayerMetatype,
    layer_metatypes.TFEmbeddingLayerMetatype,
    layer_metatypes.TFLocallyConnected1DLayerMetatype,
    layer_metatypes.TFLocallyConnected2DLayerMetatype,
    # ALL_TF_OP_METATYPES_WITH_WEIGHTS
    op_metatypes.TFConv2DOpMetatype,
    op_metatypes.TFConv3DOpMetatype,
    op_metatypes.TFMatMulOpMetatype,
    op_metatypes.TFDepthwiseConv2dNativeOpMetatype,
]

GENERAL_CONV_LAYER_METATYPES = [
    layer_metatypes.TFConv1DLayerMetatype,
    layer_metatypes.TFConv2DLayerMetatype,
    layer_metatypes.TFConv3DLayerMetatype,
    layer_metatypes.TFDepthwiseConv1DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv3DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DLayerMetatype,
    layer_metatypes.TFConv1DTransposeLayerMetatype,
    layer_metatypes.TFConv2DTransposeLayerMetatype,
    layer_metatypes.TFConv3DTransposeLayerMetatype,
    # GENERAL_CONV_TF_OP_METATYPES
    op_metatypes.TFConv2DOpMetatype,
    op_metatypes.TFConv3DOpMetatype,
    op_metatypes.TFDepthwiseConv2dNativeOpMetatype,
]

DEPTHWISE_CONV_LAYER_METATYPES = [
    layer_metatypes.TFDepthwiseConv1DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv3DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DLayerMetatype,
    # DEPTHWISE_CONV_TF_OP_METATYPES
    op_metatypes.TFDepthwiseConv2dNativeOpMetatype,
]

DECONV_LAYER_METATYPES = [
    layer_metatypes.TFConv1DTransposeLayerMetatype,
    layer_metatypes.TFConv2DTransposeLayerMetatype,
    layer_metatypes.TFConv3DTransposeLayerMetatype,
]

LINEAR_LAYER_METATYPES = [
    layer_metatypes.TFDenseLayerMetatype,
    # LINEAR_TF_OP_METATYPES
    op_metatypes.TFMatMulOpMetatype,
]

NORMALIZATION_LAYER_METATYPES = [
    layer_metatypes.TFBatchNormalizationLayerMetatype,
    layer_metatypes.TFLayerNormalizationLayerMetatype,
    # NORMALIZATION_TF_OP_METATYPES
    op_metatypes.TFFusedBatchNormV3OpMetatype,
]

LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT = [
    layer_metatypes.TFCropping1DLayerMetatype,
    layer_metatypes.TFCropping2DLayerMetatype,
    layer_metatypes.TFCropping3DLayerMetatype,
    layer_metatypes.TFFlattenLayerMetatype,
    layer_metatypes.TFGlobalMaxPooling1DLayerMetatype,
    layer_metatypes.TFGlobalMaxPooling2DLayerMetatype,
    layer_metatypes.TFGlobalMaxPooling3DLayerMetatype,
    layer_metatypes.TFMaxPooling1DLayerMetatype,
    layer_metatypes.TFMaxPooling2DLayerMetatype,
    layer_metatypes.TFMaxPooling3DLayerMetatype,
    layer_metatypes.TFRepeatVectorLayerMetatype,
    layer_metatypes.TFReshapeLayerMetatype,
    layer_metatypes.TFZeroPadding1DLayerMetatype,
    layer_metatypes.TFZeroPadding2DLayerMetatype,
    layer_metatypes.TFZeroPadding3DLayerMetatype,
    layer_metatypes.TFDropoutLayerMetatype,
    layer_metatypes.TFSlicingOpLambdaMetatype,
    layer_metatypes.TFPermuteLayerMetatype,
    # TF_OP_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT
    op_metatypes.TFIdentityOpMetatype,
    op_metatypes.TFPackOpMetatype,
    op_metatypes.TFPadOpMetatype,
    op_metatypes.TFStridedSliceOpMetatype,
    op_metatypes.TFSliceOpMetatype,
    op_metatypes.TFReshapeOpMetatype,
    op_metatypes.TFShapeOpMetatype,
    op_metatypes.TFMaxOpMetatype,
    op_metatypes.TFMaxPoolOpMetatype,
    op_metatypes.TFExpandDimsOpMetatype,
    op_metatypes.TFSqueezeOpMetatype,
    op_metatypes.TFGatherOpMetatype,
    op_metatypes.TFMaxPool3DOpMetatype,
    op_metatypes.TFTileOpMetatype,
    op_metatypes.TFSplitOpMetatype,
    op_metatypes.TFTransposeOpMetatype,
    # TFReluOpMetatype and TFReLULayerMetatype aren't considered to be QUANTIZATION_AGNOSTIC, because:
    # 1. Runtime doesn't provide performance benefits by quantizing the stand-alone RELU's (ticket: 59548)
    # 2. It's frequently better for the end accuracy to have quantizers set up after the RELU
    # so that the input distribution to the quantizer is non-negative
    # and we can therefore have better quantization resolution while preserving the original dynamic range
]

LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_CONCAT_INPUTS = [
    layer_metatypes.TFConcatenateLayerMetatype,
    op_metatypes.TFConcatOpMetatype,
]

LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_INPUTS = [
    op_metatypes.TFMaximumOpMetatype,
    op_metatypes.TFMinimumOpMetatype,
]

LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_OUTPUTS = [
    op_metatypes.TFSplitOpMetatype,
    op_metatypes.TFUnPackOpMetatype,
]

ELEMENTWISE_LAYER_METATYPES = [
    layer_metatypes.TFAddLayerMetatype,
    layer_metatypes.TFMultiplyLayerMetatype,
    layer_metatypes.TFRescalingLayerMetatype,
    # ELEMENTWISE_TF_OP_METATYPES
    op_metatypes.TFAddOpMetatype,
    op_metatypes.TFMulOpMetatype,
    op_metatypes.TFBiasAddOpMetatype,
    op_metatypes.TFGreaterOpMetatype,
    op_metatypes.TFNegOpMetatype,
    op_metatypes.TFSubOpMetatype,
    op_metatypes.TFFloorDivOpMetatype,
    op_metatypes.TFMaximumOpMetatype,
    op_metatypes.TFMinimumOpMetatype,
]

RESHAPE_METATYPES = [
    layer_metatypes.TFReshapeLayerMetatype,
    layer_metatypes.TFFlattenLayerMetatype,
    op_metatypes.TFReshapeOpMetatype,
]

DIMENSION_PERMUTATION_METATYPES = [
    op_metatypes.TFTransposeOpMetatype,
    layer_metatypes.TFPermuteLayerMetatype,
]

CAST_METATYPES = [
    op_metatypes.TFCastOpMetatype,
]


def get_operator_metatypes() -> List[Type[OperatorMetatype]]:
    keras_metatypes_list = list(layer_metatypes.KERAS_LAYER_METATYPES.registry_dict.values())
    tf_metatypes_list = list(op_metatypes.TF_OPERATION_METATYPES.registry_dict.values())
    return list(
        set(
            keras_metatypes_list
            + tf_metatypes_list
            + list(INPUT_NOOP_METATYPES.registry_dict.values())
            + list(OUTPUT_NOOP_METATYPES.registry_dict.values())
            + list(NOOP_METATYPES.registry_dict.values())
        )
    )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Structures and functions for passing advanced parameters to NNCF post-training quantization APIs.
"""

import sys
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from dataclasses import is_dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union

import nncf
from nncf.common.quantization.quantizer_propagation.structs import QuantizerPropagationRule
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.utils.api_marker import api
from nncf.parameters import StrEnum
from nncf.quantization.range_estimator import AggregatorType
from nncf.quantization.range_estimator import RangeEstimatorParameters
from nncf.quantization.range_estimator import StatisticsType


@api(canonical_alias="nncf.OverflowFix")
class OverflowFix(StrEnum):
    """
    This option controls whether to apply the overflow issue fix for the 8-bit
    quantization.

    8-bit instructions of older Intel CPU generations (based on SSE, AVX-2, and AVX-512
    instruction sets) suffer from the so-called saturation (overflow) issue: in some
    configurations, the output does not fit into an intermediate buffer and has to be
    clamped. This can lead to an accuracy drop on the aforementioned architectures.
    The fix set to use only half a quantization range to avoid overflow for specific
    operations.

    If you are going to infer the quantized model on the architectures with AVX-2, and
    AVX-512 instruction sets, we recommend using FIRST_LAYER option as lower aggressive
    fix of the overflow issue. If you still face significant accuracy drop, try using
    ENABLE, but this may get worse the accuracy.

    :param ENABLE: All weights of all types of Convolutions and MatMul operations
        are be quantized using a half of the 8-bit quantization range.
    :param FIRST_LAYER: Weights of the first Convolutions of each model inputs
        are quantized using a half of the 8-bit quantization range.
    :param DISABLE: All weights are quantized using the full 8-bit quantization range.
    """

    ENABLE = "enable"
    FIRST_LAYER = "first_layer_only"
    DISABLE = "disable"


@api()
class FP8Type(StrEnum):
    """
    Defines FP8 special types (https://arxiv.org/pdf/2209.05433.pdf).

    :param E4M3: Mode with 4-bit exponent and 3-bit mantissa.
    :param E5M2: Mode with 5-bit exponent and 2-bit mantissa.

    """

    E4M3 = "f8e4m3"
    E5M2 = "f8e5m2"


@api()
@dataclass
class QuantizationParameters:
    """
    Contains quantization parameters for weights or activations.

    :param num_bits: The number of bits to use for quantization.
    :type num_bits: Optional[int]
    :param mode: The quantization mode to use, such as 'symmetric', 'asymmetric', etc.
    :type mode: nncf.common.quantization.structs.QuantizationMode
    :param signedness_to_force: Whether to force the weights or activations to be
        signed (True), unsigned (False)
    :type signedness_to_force: Optional[bool]
    :param per_channel: True if per-channel quantization is used, and False if
        per-tensor quantization is used.
    :type per_channel: Optional[bool]
    :param narrow_range: Whether to use a narrow quantization range.

        If False, then the input will be quantized into quantization range

        * [0; 2^num_bits - 1] for unsigned quantization and
        * [-2^(num_bits - 1); 2^(num_bits - 1) - 1] for signed quantization

        If True, then the ranges would be:

        * [0; 2^num_bits - 2] for unsigned quantization and
        * [-2^(num_bits - 1) + 1; 2^(num_bits - 1) - 1] for signed quantization
    :type narrow_range: Optional[bool]
    """

    num_bits: Optional[int] = None
    mode: Optional[QuantizationMode] = None
    signedness_to_force: Optional[bool] = None
    per_channel: Optional[bool] = None
    narrow_range: Optional[bool] = None


@api()
@dataclass
class FP8QuantizationParameters:
    """
    Contains convert parameters for weights or activations.

    :param destination_type: Currently contains E4M3 or E5M2 for FP8 precision.
    :type destination_type: FP8Type
    """

    destination_type: Optional[FP8Type] = None


@api()
@dataclass
class AdvancedBiasCorrectionParameters:
    """
    Contains advanced parameters for fine-tuning bias correction algorithm.

    :param apply_for_all_nodes: Whether to apply the correction to all nodes in the
        model, or only to nodes that have a bias.
    :type apply_for_all_nodes: bool
    :param threshold: The threshold value determines the maximum bias correction value.
        The bias correction are skipped If the value is higher than threshold.
    :type threshold: Optional[float]
    """

    apply_for_all_nodes: bool = False
    threshold: Optional[float] = None


@api()
@dataclass
class AdvancedSmoothQuantParameters:
    """
    Contains advanced alpha parameters for SmoothQuant algorithm.
    It regulates the calculation of the smooth scale for different node types.
    A negative value switches off the algorithm for current node type. In case of inaccurate results,
    this parameter may be adjusted in the range from 0 to 1 or set -1 to disable SmoothQuant algorithm.

    :param convolution: Whether to apply smoothing for Convolution layers.
    :type convolution: float
    :param matmul: Whether to apply smoothing for MatMul layers.
    :type matmul: float
    """

    convolution: float = -1
    matmul: float = 0.95


class RestoreMode(Enum):
    """
    Specifies how to revert operations to their original precision.

    :param ACTIVATIONS_AND_WEIGHTS: Operations will be reverted to floating-point precision.
    :param ONLY_ACTIVATIONS: Operations with weights will be reverted to representation with int8 weights,
        while all other operations will revert to floating-point precision.
    """

    ACTIVATIONS_AND_WEIGHTS = "activations_and_weights"
    ONLY_ACTIVATIONS = "only_activations"


@api()
@dataclass
class AdvancedQuantizationParameters:
    """
    Contains advanced parameters for fine-tuning quantization algorithm.

    :param overflow_fix: This option controls whether to apply the overflow issue fix
        for the 8-bit quantization.
    :type overflow_fix: Optional[nncf.quantization.advanced_parameters.OverflowFix]
    :param quantize_outputs: Whether to insert additional quantizers right before each
        of the model outputs.
    :type quantize_outputs: bool
    :param inplace_statistics: Defines whether to calculate quantizers statistics by
        backend graph operations or by default Python implementation, defaults to True.
    :type inplace_statistics: bool
    :param disable_channel_alignment: Whether to disable the channel alignment.
    :type disable_channel_alignment: bool
    :param disable_bias_correction: Whether to disable the bias correction.
    :type disable_bias_correction: bool
    :param batchwise_statistics: Determines whether quantizer statistics should be calculated
        for each item of the batch or for the entire batch, default is None.
        "None" means that if torch.DataLoader or tensorflow.Dataset was passed as a data source for
        the calibration dataset, then in case batch_size of the data source > 1 batchwise_statistics sets to True,
        otherwise sets to False.
    :type batchwise_statistics: Optional[bool]
    :param quantizer_propagation_rule: An instance of the `QuantizerPropagationRule` enum that
        specifies how quantizers should be propagated and merged across branching nodes in the
        model's computational graph. The strategies are as follows:
        - DO_NOT_MERGE_BRANCHES: No merging of quantization parameters across branches.
        - MERGE_IF_ALL_BRANCHES_SAME : Merge only if all branch quantization configurations are identical.
        - MERGE_WITH_POTENTIAL_REQUANTIZATION: Merge common configurations and allow for requantization
        on branches with additional options.
        - MERGE_ALL_IN_ONE: Attempt to merge into a single global quantization configuration
        if possible given hardware constraints.
        MERGE_ALL_IN_ONE is a default value.
    :type quantizer_propagation_rule: QuantizerPropagationRule
    :param activations_quantization_params: Quantization parameters for activations.
    :type activations_quantization_params: nncf.quantization.advanced_parameters.QuantizationParameters
    :param weights_quantization_params: Quantization parameters for weights.
    :type weights_quantization_params: nncf.quantization.advanced_parameters.QuantizationParameters
    :param activations_range_estimator_params: Range estimator parameters for activations.
    :type activations_range_estimator_params: nncf.quantization.range_estimator.RangeEstimatorParameters
    :param weights_range_estimator_params: Range estimator parameters for weights.
    :type weights_range_estimator_params: nncf.quantization.range_estimator.RangeEstimatorParameters
    :param bias_correction_params: Advanced bias correction parameters.
    :type bias_correction_params: nncf.quantization.advanced_parameters.AdvancedBiasCorrectionParameters
    :param smooth_quant_alphas: SmoothQuant-related parameters mapping.
        It regulates the calculation of the smooth scale. The default value stored in AdvancedSmoothQuantParameters.
        A negative value for each field switches off type smoothing. In case of inaccurate results,
        fields may be adjusted in the range from 0 to 1 or set -1 to disable smoothing for type.
    :type smooth_quant_alphas: nncf.quantization.advanced_parameters.AdvancedSmoothQuantParameters
    :param smooth_quant_alpha: Deprecated SmoothQuant-related parameter.
    :type smooth_quant_alpha: float
    :param backend_params: Backend-specific parameters.
    :type backend_params: Dict[str, Any]
    """

    # General parameters
    overflow_fix: Optional[OverflowFix] = None
    quantize_outputs: bool = False
    inplace_statistics: bool = True
    disable_channel_alignment: bool = True
    disable_bias_correction: bool = False
    batchwise_statistics: Optional[bool] = None

    # Advanced Quantization parameters
    activations_quantization_params: Optional[Union[QuantizationParameters, FP8QuantizationParameters]] = None
    weights_quantization_params: Optional[Union[QuantizationParameters, FP8QuantizationParameters]] = None
    quantizer_propagation_rule: QuantizerPropagationRule = QuantizerPropagationRule.MERGE_ALL_IN_ONE

    # Range estimator parameters
    activations_range_estimator_params: RangeEstimatorParameters = field(default_factory=RangeEstimatorParameters)
    weights_range_estimator_params: RangeEstimatorParameters = field(default_factory=RangeEstimatorParameters)

    # Advanced BiasCorrection algorithm parameters
    bias_correction_params: AdvancedBiasCorrectionParameters = field(default_factory=AdvancedBiasCorrectionParameters)

    # Advanced SmoothQuant algorithm parameters
    smooth_quant_alphas: AdvancedSmoothQuantParameters = field(default_factory=AdvancedSmoothQuantParameters)
    # Deprecated parameter
    smooth_quant_alpha: Optional[float] = None

    # Backend specific parameters
    backend_params: Dict[str, Any] = field(default_factory=dict)


@api()
@dataclass
class AdvancedAWQParameters:
    """
    Contains advanced parameters for AWQ algorithm.

    :param subset_size: The number of samples for AWQ.
    :type subset_size: int
    :param percent_to_apply: The percent of outliers for correction.
    :type percent_to_apply: float
    :param alpha_min: Minimum value of smoothness parameter for grid search.
    :type alpha_min: float
    :param alpha_max: Maximal value of smoothness parameter for grid search.
    :type alpha_max: float
    :param steps: The number of the steps in grid search.
    :type steps: int
    """

    subset_size: int = 32
    percent_to_apply: float = 0.002
    alpha_min: float = 0.0
    alpha_max: float = 1.0
    steps: int = 100


@api()
@dataclass
class AdvancedScaleEstimationParameters:
    """
    Contains advanced parameters for scale estimation algorithm.

    :param subset_size: The number of samples for scale estimation.
    :type subset_size: int
    :param initial_steps: The number of the steps for absmax scale rectification.
    :type initial_steps: int
    :param scale_steps: The number of the steps for grid search scale rectification
        from 1.0 to 1.0 - 0.05 * scale_step.
    :type scale_steps: int
    :param weight_penalty: coefficient for penalty between fp and compressed weights. If -1 then doesn't apply.
    :type weight_penalty: float
    """

    subset_size: int = 64
    initial_steps: int = 5
    scale_steps: int = 5
    weight_penalty: float = -1.0


@api()
@dataclass
class AdvancedGPTQParameters:
    """
    Contains advanced parameters for GPTQ algorithm.

    :param damp_percent: The percent of the average Hessian diagonal to use for dampening,
        recommended value is 0.1.
    :type damp_percent: float
    :param block_size: The size of the blocks used during quantization. Defaults to 128.
    :type block_size: int
    :param subset_size: Number of data samples to calculate Hessian. Defaults to 128.
    :type subset_size: int
    """

    damp_percent: float = 0.1
    block_size: int = 128
    subset_size: int = 128


@api()
@dataclass
class AdvancedLoraCorrectionParameters:
    """
    Contains advanced parameters for lora correction algorithm.

    :param adapter_rank: rank of lora adapters. Defaults to 16.
    :type adapter_rank: int
    :param num_iterations: number of correction iterations. Defaults to 3.
    :type num_iterations: int
    :param apply_regularization: Whether to add a regularization during the correction process. Defaults to True.
        Helpful for big rank values to avoid overfitting.
    :type apply_regularization: bool
    :param subset_size: Number of data samples for lora correction algorithm. Defaults to 128.
    :type subset_size: int
    :param use_int8_adapters: Whether to 8-bit quantize lora adapters, otherwise they kept in the original weights
        precision. Defaults to True.
    :type use_int8_adapters: bool
    """

    adapter_rank: int = 8
    num_iterations: int = 3
    apply_regularization: bool = True
    subset_size: int = 128
    use_int8_adapters: bool = True


@api()
@dataclass
class AdvancedCompressionParameters:
    """
    Contains advanced parameters for compression algorithms.

    :param statistics_path: Directory path to dump statistics.
    :type statistics_path: str
    :param awq_params: Advanced parameters for AWQ algorithm.
    :type awq_params: AdvancedAWQParameters
    :param scale_estimation_params: Advanced parameters for scale estimation algorithm.
    :type scale_estimation_params: AdvancedScaleEstimationParameters
    """

    statistics_path: Optional[str] = None
    # Advanced AWQ algorithm parameters
    awq_params: AdvancedAWQParameters = field(default_factory=AdvancedAWQParameters)

    # Advanced scale estimation algorithm parameters
    scale_estimation_params: AdvancedScaleEstimationParameters = field(
        default_factory=AdvancedScaleEstimationParameters
    )

    # Advanced GPTQ algorithm parameters
    gptq_params: AdvancedGPTQParameters = field(default_factory=AdvancedGPTQParameters)

    # Advanced Lora Correction algorithm parameters
    lora_correction_params: AdvancedLoraCorrectionParameters = field(default_factory=AdvancedLoraCorrectionParameters)


@api()
@dataclass
class AdvancedAccuracyRestorerParameters:
    """
    Contains advanced parameters for fine-tuning the accuracy restorer algorithm.

    :param max_num_iterations: The maximum number of iterations of the algorithm.
        In other words, the maximum number of layers that may be reverted back to
        floating-point precision. By default, it is limited by the overall number of
        quantized layers.
    :type max_num_iterations: int
    :param tune_hyperparams: Whether to tune of quantization parameters as a
        preliminary step before reverting layers back to the floating-point precision.
        It can bring an additional boost in performance and accuracy, at the cost of
        increased overall quantization time. The default value is `False`.
    :type tune_hyperparams: int
    :param ranking_subset_size: Size of a subset that is used to rank layers by their
        contribution to the accuracy drop.
    :type ranking_subset_size: Optional[int]
    :param num_ranking_workers: The number of parallel workers that are used to rank
        quantization operations.
    :type num_ranking_workers: Optional[int]
    :param intermediate_model_dir: Path to the folder where the model, which was fully
        quantized with initial parameters, should be saved.
    :type intermediate_model_dir: Optional[str]
    :param restore_mode: Specifies how to revert operations to their original precision.
    :type restore_mode: RestoreMode
    """

    max_num_iterations: int = sys.maxsize
    tune_hyperparams: bool = False
    ranking_subset_size: Optional[int] = None
    num_ranking_workers: Optional[int] = None
    intermediate_model_dir: Optional[str] = None
    restore_mode: RestoreMode = RestoreMode.ACTIVATIONS_AND_WEIGHTS


def changes_asdict(params: Any) -> Dict[str, Any]:
    """
    Returns non None fields as dict

    :param params: A dataclass instance
    :return: A dict with non None fields
    """
    changes = {}
    for f in fields(params):
        value = getattr(params, f.name)
        if value is not None:
            changes[f.name] = value
    return changes


def convert_to_dict_recursively(params: Any) -> Dict[str, Any]:
    """
    Converts dataclass to dict recursively

    :param params: A dataclass instance
    :return: A dataclass as dict
    """
    if params is None:
        return {}

    result = {}
    for f in fields(params):
        value = getattr(params, f.name)
        if is_dataclass(value):
            result[f.name] = convert_to_dict_recursively(value)
        elif isinstance(value, Enum):
            result[f.name] = value.value
        else:
            result[f.name] = value

    return result


def convert_quantization_parameters_to_dict(params: Optional[QuantizationParameters]) -> Dict[str, Any]:
    """
    Converts quantization parameters to the dict in the legacy format

    :param params: Quantization parameters
    :return: Quantization parameters as dict in the legacy format
    """
    result: Dict[str, Any] = {}
    if params is not None:
        if params.num_bits is not None:
            result["bits"] = params.num_bits
        if params.mode is not None:
            result["mode"] = params.mode
        if params.signedness_to_force is not None:
            result["signed"] = params.signedness_to_force
        if params.per_channel is not None:
            result["per_channel"] = params.per_channel
        if params.narrow_range is not None:
            raise nncf.ParameterNotSupportedError("narrow_range parameter is not supported in the legacy format")
    return result


def convert_range_estimator_parameters_to_dict(params: RangeEstimatorParameters) -> Dict[str, Any]:
    """
    Converts range estimator parameters to the dict in the legacy format

    :param params: Range estimator parameters
    :return: range estimator parameters as dict in the legacy format
    """
    if params.min.clipping_value is not None or params.max.clipping_value is not None:
        raise nncf.ParameterNotSupportedError("clipping_value parameter is not supported in the legacy format")

    result: Dict[str, Any] = {}
    if (
        params.min.statistics_type == StatisticsType.MIN
        and params.min.aggregator_type == AggregatorType.MIN
        and params.max.statistics_type == StatisticsType.MAX
        and params.max.aggregator_type == AggregatorType.MAX
    ):
        result["type"] = "mixed_min_max"
    elif (
        params.min.statistics_type == StatisticsType.MIN
        and params.min.aggregator_type == AggregatorType.MEAN
        and params.max.statistics_type == StatisticsType.MAX
        and params.max.aggregator_type == AggregatorType.MEAN
    ):
        result["type"] = "mean_min_max"
    elif (
        params.min.statistics_type == StatisticsType.QUANTILE
        and params.min.aggregator_type == AggregatorType.MEAN
        and params.max.statistics_type == StatisticsType.QUANTILE
        and params.max.aggregator_type == AggregatorType.MEAN
    ):
        result["type"] = "mean_percentile"
        result["params"] = {
            "min_percentile": 1 - params.min.quantile_outlier_prob,
            "max_percentile": 1 - params.max.quantile_outlier_prob,
        }
    elif (
        params.min.statistics_type is None
        and params.min.aggregator_type is None
        and params.max.statistics_type is None
        and params.max.aggregator_type is None
    ):
        return {}
    else:
        raise nncf.ParameterNotSupportedError(
            f"The following range estimator parameters are not supported: {str(params)}"
        )

    return result


def apply_advanced_parameters_to_config(
    config: Dict[str, Any], params: AdvancedQuantizationParameters
) -> Dict[str, Any]:
    """
    Apply advanced parameters to the config in the legacy format

    :param config: NNCF config in legacy format
    :param params: Advanced quantization parameters
    :return: advanced quantization parameters as dict in the legacy format
    """
    config["overflow_fix"] = params.overflow_fix if params.overflow_fix is None else params.overflow_fix.value
    config["quantize_outputs"] = params.quantize_outputs

    if params.disable_bias_correction:
        initializer = config.get("initializer", {})
        initializer["batchnorm_adaptation"] = {"num_bn_adaptation_samples": 0}
        config["initializer"] = initializer

    if isinstance(params.activations_quantization_params, QuantizationParameters):
        activations_config = convert_quantization_parameters_to_dict(params.activations_quantization_params)
        if activations_config:
            config["activations"] = activations_config

    if isinstance(params.weights_quantization_params, QuantizationParameters):
        weights_config = convert_quantization_parameters_to_dict(params.weights_quantization_params)
        if weights_config:
            config["weights"] = weights_config

    activations_init_range_config = convert_range_estimator_parameters_to_dict(
        params.activations_range_estimator_params
    )
    weights_init_range_config = convert_range_estimator_parameters_to_dict(params.weights_range_estimator_params)

    if activations_init_range_config or weights_init_range_config:
        initializer = config.get("initializer", {})
        init_range = initializer.get("range", {})
        global_num_init_samples = init_range.get("num_init_samples", None)
        global_range_type = init_range.get("type", None)

        activations_init_range_config["target_quantizer_group"] = "activations"
        activations_init_range_config["target_scopes"] = "{re}.*"
        if global_num_init_samples is not None:
            activations_init_range_config["num_init_samples"] = global_num_init_samples
        if "type" not in activations_init_range_config and global_range_type is not None:
            activations_init_range_config["type"] = global_range_type

        weights_init_range_config["target_quantizer_group"] = "weights"
        weights_init_range_config["target_scopes"] = "{re}.*"
        if global_num_init_samples is not None:
            weights_init_range_config["num_init_samples"] = global_num_init_samples
        if "type" not in weights_init_range_config and global_range_type is not None:
            weights_init_range_config["type"] = global_range_type

        initializer["range"] = [activations_init_range_config, weights_init_range_config]
        config["initializer"] = initializer

    if params.bias_correction_params.apply_for_all_nodes:
        raise nncf.ParameterNotSupportedError(
            "apply_for_all_nodes parameter of the BiasCorrection algorithm is not supported in the legacy format"
        )

    if params.bias_correction_params.threshold is not None:
        raise nncf.ParameterNotSupportedError(
            "threshold parameter of the BiasCorrection algorithm is not supported in the legacy format"
        )

    return config



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Any

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TransformationCommand


class CommandCreator(ABC):
    """
    Creates transformation commands to use them in the model transformer.
    """

    @staticmethod
    @abstractmethod
    def create_command_to_remove_quantizer(quantizer_node: NNCFNode) -> TransformationCommand:
        """
        Creates command to remove quantizer from the quantized model.

        :param quantizer_node: The quantizer which should be removed.
        :return: The command to remove quantizer from the quantized model.
        """

    @staticmethod
    @abstractmethod
    def create_command_to_update_bias(
        node_with_bias: NNCFNode, bias_value: Any, nncf_graph: NNCFGraph
    ) -> TransformationCommand:
        """
        Creates command to update bias value.

        :param node_with_bias: The node that corresponds to the operation with bias.
        :param bias_value: New bias value.
        :param nncf_graph: The NNCF graph.
        :return: The command to update bias value.
        """

    @staticmethod
    @abstractmethod
    def create_command_to_insert_bias(node_without_bias: NNCFNode, bias_value: Any) -> TransformationCommand:
        """
        Creates command to insert bias after given node.

        :param node_without_bias: The node that corresponds to the operation without bias.
        :param bias_value: Bias value to insert.
        :return: The command to insert bias value.
        """

    @staticmethod
    @abstractmethod
    def create_command_to_update_weight(
        node_with_weight: NNCFNode, weight_value: Any, weight_port_id: int
    ) -> TransformationCommand:
        """
        Creates command to update weight value.

        :param node_with_weight: The node that corresponds to the operation with weight.
        :param weight_value: New weight value.
        :param weight_port_id: The input port ID that corresponds to weight.
        :return: The command to update weight value.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from pathlib import Path
from typing import Dict, Optional

from nncf.common.utils.os import fail_if_symlink
from nncf.tensor import Tensor
from nncf.tensor.definitions import TensorBackend
from nncf.tensor.definitions import TensorDeviceType
from nncf.tensor.functions.dispatcher import dispatch_dict
from nncf.tensor.functions.dispatcher import get_io_backend_fn


def load_file(
    file_path: Path,
    *,
    backend: TensorBackend,
    device: Optional[TensorDeviceType] = None,
) -> Dict[str, Tensor]:
    """
    Loads a file containing tensor data and returns a dictionary of tensors.

    :param file_path: The path to the file to be loaded.
    :param backend: The backend type to determine the loading function.
    :param device: The device on which the tensor will be allocated, If device is not given,
        then the default device is determined by backend.
    :return: A dictionary where the keys are tensor names and the values are Tensor objects.
    """
    fail_if_symlink(file_path)
    loaded_dict = get_io_backend_fn("load_file", backend)(file_path, device=device)
    return {key: Tensor(val) for key, val in loaded_dict.items()}


@functools.singledispatch
def save_file(
    data: Dict[str, Tensor],
    file_path: Path,
) -> None:
    """
    Saves a dictionary of tensors to a file.

    :param data: A dictionary where the keys are tensor names and the values are Tensor objects.
    :param file_path: The path to the file where the tensor data will be saved.
    """
    fail_if_symlink(file_path)
    if isinstance(data, dict):
        return dispatch_dict(save_file, data, file_path)
    raise NotImplementedError(f"Function `save_file` is not implemented for {type(data)}")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from shutil import copyfile
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar

import torch

from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmController
from nncf.api.compression import CompressionStage
from nncf.common.logging import nncf_logger
from nncf.config.structures import BNAdaptationInitArgs
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.training.base_training import BNASTrainingController
from nncf.experimental.torch.nas.bootstrapNAS.training.model_creator_helpers import (
    create_compressed_model_from_algo_names,
)
from nncf.experimental.torch.nas.bootstrapNAS.training.model_creator_helpers import resume_compression_from_state
from nncf.torch.checkpoint_loading import load_state
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import is_main_process

TModel = TypeVar("TModel")
OptimizerType = TypeVar("OptimizerType")
LRSchedulerType = TypeVar("LRSchedulerType")
TensorboardWriterType = TypeVar("TensorboardWriterType")
DataLoaderType = TypeVar("DataLoaderType")
TrainEpochFnType = Callable[
    [
        DataLoaderType,
        TModel,
        CompressionAlgorithmController,
        int,
        OptimizerType,
    ],
    None,
]
ValFnType = Callable[[TModel, DataLoaderType], Tuple[float, float, float]]


class EBTrainAlgoStateNames:
    MODEL_STATE = "model_state"
    EPOCH = "epoch"
    SUPERNET_ACC1 = "acc1"
    SUPERNET_BEST_ACC1 = "best_acc1"
    MIN_SUBNET_ACC1 = "min_subnet_acc1"
    MIN_SUBNET_BEST_ACC1 = "min_subnet_best_acc1"
    OPTIMIZER = "optimizer"
    TRAINING_ALGO_STATE = "training_algo_state"


class EpochBasedTrainingAlgorithm:
    """
    Algorithm for training supernet by using a train function for a single epoch. In contrast, there can be step-based
    algorithm that uses a train function for a single training step.
    """

    _state_names = EBTrainAlgoStateNames

    def __init__(
        self,
        nncf_network: NNCFNetwork,
        training_ctrl: BNASTrainingController,
        checkpoint: Optional[Dict[str, Any]] = None,
    ):
        """
        Initializes the training algorithm

        :param nncf_network: it's supposed to be a model with elasticity ops,
        controlled by elasticity ctrl, which is owned by training ctrl
        :param training_ctrl: controller of the training algorithm (by default ProgressiveShrinkingController)
        :param checkpoint: data to restore state of the training algorithm
        """
        self._model = nncf_network
        self._training_ctrl = training_ctrl
        self._start_epoch = 0
        self._supernet_best_acc1 = 0
        self._min_subnet_best_acc1 = 0
        self._optimizer = None
        self._optimizer_state = None
        if checkpoint is not None:
            resuming_model_state_dict = checkpoint[self._state_names.MODEL_STATE]
            load_state(self._model, resuming_model_state_dict, is_resume=True)
            self._optimizer_state = checkpoint[self._state_names.OPTIMIZER]
            self._start_epoch = checkpoint[self._state_names.EPOCH]
            self._supernet_best_acc1 = checkpoint[self._state_names.SUPERNET_BEST_ACC1]
            self._min_subnet_best_acc1 = checkpoint[self._state_names.MIN_SUBNET_BEST_ACC1]

    @property
    def elasticity_ctrl(self) -> ElasticityController:
        """
        :return: elasticity controller
        """
        return self._training_ctrl.elasticity_controller

    def run(
        self,
        train_epoch_fn: TrainEpochFnType,
        train_loader: DataLoaderType,
        val_fn: ValFnType,
        val_loader: DataLoaderType,
        optimizer: OptimizerType,
        checkpoint_save_dir: str,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
        train_iters: Optional[float] = None,
    ) -> Tuple[NNCFNetwork, ElasticityController]:
        """
        Implements a training loop for supernet training.

        :param train_epoch_fn: a method to fine-tune the model for a single epoch
        :param train_loader: data loader for training
        :param lr_scheduler: scheduler for learning rate
        :param val_fn: a method to evaluate the model on the validation dataset
        :param val_loader: data loader for validation
        :param optimizer: a training optimizer
        :param checkpoint_save_dir: path to a directory for saving checkpoints
        :param tensorboard_writer: The tensorboard object to be used for logging.
        :return: the fine-tuned model and elasticity controller
        """

        if train_iters is None:
            train_iters = len(train_loader)
        self._training_ctrl.set_training_lr_scheduler_args(optimizer, train_iters)  # len(train_loader))

        if self._optimizer_state is not None:
            optimizer.load_state_dict(self._optimizer_state)

        self._optimizer = optimizer

        log_validation_info = True
        if tensorboard_writer is None:
            try:
                from torch.utils.tensorboard import SummaryWriter

                tensorboard_writer = SummaryWriter(checkpoint_save_dir)
                # log compression config to tensorboard
            except ModuleNotFoundError:
                nncf_logger.warning(
                    "Tensorboard installation not found! Install tensorboard Python package "
                    "in order for BootstrapNAS tensorboard data to be dumped"
                )

        total_num_epochs = self._training_ctrl.get_total_num_epochs()

        best_compression_stage = CompressionStage.UNCOMPRESSED
        for epoch in range(self._start_epoch, total_num_epochs):
            self._training_ctrl.scheduler.epoch_step()

            # TODO(nlyalyus): support DDP (ticket 56849)
            # if config.distributed:
            #     train_sampler.set_epoch(epoch)

            # train for one epoch
            train_epoch_fn(train_loader, self._model, self._training_ctrl, epoch, optimizer)

            compression_stage = self._training_ctrl.compression_stage()

            self._training_ctrl.multi_elasticity_handler.activate_minimum_subnet()
            min_subnet_acc1, acc5, loss = self._validate_subnet(val_fn, val_loader)
            if log_validation_info:
                nncf_logger.info(
                    f"* Acc@1 {min_subnet_acc1:.3f} Acc@5 {acc5:.3f} "
                    f"for Minimal SubNet={self._training_ctrl.multi_elasticity_handler.get_active_config()}"
                )
            if is_main_process() and log_validation_info:
                tensorboard_writer.add_scalar("val/min_subnet_loss", loss, len(val_loader) * epoch)
                tensorboard_writer.add_scalar("val/min_subnet_top1", min_subnet_acc1, len(val_loader) * epoch)
                tensorboard_writer.add_scalar("val/min_subnet_top5", acc5, len(val_loader) * epoch)
            min_subnet_best_acc1 = self._define_best_accuracy(
                min_subnet_acc1, self._min_subnet_best_acc1, compression_stage, best_compression_stage
            )

            self._training_ctrl.multi_elasticity_handler.activate_supernet()
            supernet_acc1, acc5, loss = self._validate_subnet(val_fn, val_loader)
            if log_validation_info:
                nncf_logger.info(
                    f"* Acc@1 {supernet_acc1:.3f} Acc@5 {acc5:.3f} "
                    f"of SuperNet={self._training_ctrl.multi_elasticity_handler.get_active_config()}"
                )

            if is_main_process() and log_validation_info:
                tensorboard_writer.add_scalar("val/supernet_loss", loss, len(val_loader) * epoch)
                tensorboard_writer.add_scalar("val/supernet_top1", supernet_acc1, len(val_loader) * epoch)
                tensorboard_writer.add_scalar("val/supernet_top5", acc5, len(val_loader) * epoch)
            supernet_best_acc1 = self._define_best_accuracy(
                supernet_acc1, self._supernet_best_acc1, compression_stage, best_compression_stage
            )

            best_compression_stage = max(compression_stage, best_compression_stage)

            checkpoint_path = Path(checkpoint_save_dir, "supernet_last.pth")
            checkpoint = {
                self._state_names.EPOCH: epoch + 1,
                self._state_names.MODEL_STATE: self._model.state_dict(),
                self._state_names.SUPERNET_BEST_ACC1: supernet_best_acc1,
                self._state_names.SUPERNET_ACC1: supernet_acc1,
                self._state_names.MIN_SUBNET_BEST_ACC1: min_subnet_best_acc1,
                self._state_names.MIN_SUBNET_ACC1: min_subnet_acc1,
                self._state_names.OPTIMIZER: optimizer.state_dict(),
                self._state_names.TRAINING_ALGO_STATE: self._training_ctrl.get_compression_state(),
            }
            torch.save(checkpoint, checkpoint_path)

            if compression_stage == CompressionStage.FULLY_COMPRESSED and supernet_best_acc1 == supernet_acc1:
                best_path = Path(checkpoint_save_dir) / "supernet_best.pth"
                copyfile(checkpoint_path, best_path)

            # Backup elasticity state and model weight to directly restore from it in a separate search sample
            elasticity_path = Path(checkpoint_save_dir) / "last_elasticity.pth"
            elasticity_state = self._training_ctrl.elasticity_controller.get_compression_state()
            model_path = Path(checkpoint_save_dir) / "last_model_weights.pth"
            model_state = self._model.state_dict()
            torch.save(elasticity_state, elasticity_path)
            torch.save(model_state, model_path)

        return self._model, self.elasticity_ctrl

    @classmethod
    def from_config(cls, nncf_network: NNCFNetwork, nncf_config: NNCFConfig) -> "EpochBasedTrainingAlgorithm":
        """
        Creates the training algorithm from a config by a given empty NNCFNetwork.

        :param nncf_network: empty NNCFNetwork
        :param nncf_config: parameters of the training algorithm
        :return: the training algorithm
        """
        algo_name = nncf_config.get("bootstrapNAS", {}).get("training", {}).get("algorithm", "progressive_shrinking")
        training_ctrl, model = create_compressed_model_from_algo_names(
            nncf_network, nncf_config, algo_names=[algo_name]
        )
        return EpochBasedTrainingAlgorithm(model, training_ctrl)

    @classmethod
    def from_checkpoint(
        cls, nncf_network: NNCFNetwork, bn_adapt_args: BNAdaptationInitArgs, resuming_checkpoint_path: str
    ) -> "EpochBasedTrainingAlgorithm":
        """
        Resumes the training algorithm from a checkpoint by a given empty NNCFNetwork and BN adaptation arguments only,
        config is not involved.

        :param nncf_network: empty NNCFNetwork
        :param bn_adapt_args: arguments for batchnorm statistics adaptation algorithm
        :param resuming_checkpoint_path: path to the resuming checkpoint
        :return: the training algorithm
        """
        if not Path(resuming_checkpoint_path).is_file():
            raise FileNotFoundError("no checkpoint found at '{}'".format(resuming_checkpoint_path))
        nncf_logger.info(f"=> loading checkpoint '{resuming_checkpoint_path}'")
        checkpoint = torch.load(resuming_checkpoint_path, map_location="cpu")

        training_state = checkpoint[cls._state_names.TRAINING_ALGO_STATE]
        nncf_config = NNCFConfig()
        nncf_config.register_extra_structs([bn_adapt_args])
        model, training_ctrl = resume_compression_from_state(nncf_network, training_state, nncf_config)
        return EpochBasedTrainingAlgorithm(model, training_ctrl, checkpoint)

    @staticmethod
    def _define_best_accuracy(
        acc1: float, best_acc1: float, compression_stage: CompressionStage, best_compression_stage: CompressionStage
    ) -> float:
        """
        The best accuracy value should be considered not only by value, but also per each stage.
        Currently, there are two stages in NAS algo only: PARTIALLY_COMPRESSED and FULLY_COMPRESSED.
        It's aligned with Compression Algorithm stages, but ideally number of stages should correspond to number of
        scheduler stages (usually more than two).
        The last means that scheduler is activated the last stage.
        """
        is_best_by_accuracy = acc1 > best_acc1 and compression_stage == best_compression_stage
        is_best = is_best_by_accuracy or compression_stage > best_compression_stage
        if is_best:
            best_acc1 = acc1
        return best_acc1

    def _validate_subnet(self, val_fn: ValFnType, val_loader: DataLoaderType):
        self._training_ctrl.prepare_for_validation()
        return val_fn(self._model, val_loader)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.common.tensor_statistics.statistics import MedianMADTensorStatistic
from nncf.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.common.tensor_statistics.statistics import PercentileTensorStatistic
from nncf.common.tensor_statistics.statistics import TensorStatistic


class TFMinMaxTensorStatistic(MinMaxTensorStatistic):
    @staticmethod
    def tensor_eq(tensor1: tf.Tensor, tensor2: tf.Tensor, rtol=1e-6) -> bool:
        return bool(tf.experimental.numpy.allclose(tensor1, tensor2, rtol=rtol))


class TFMedianMADTensorStatistic(MedianMADTensorStatistic):
    @staticmethod
    def tensor_eq(tensor1: tf.Tensor, tensor2: tf.Tensor, rtol=1e-6) -> bool:
        return bool(tf.experimental.numpy.allclose(tensor1, tensor2, rtol=rtol))


class TFPercentileTensorStatistic(PercentileTensorStatistic):
    @staticmethod
    def tensor_eq(tensor1: tf.Tensor, tensor2: tf.Tensor, rtol=1e-6) -> bool:
        return bool(tf.experimental.numpy.allclose(tensor1, tensor2, rtol=rtol))


def tf_convert_stat_to_min_max_tensor_stat(statistic: TensorStatistic) -> TFMinMaxTensorStatistic:
    if isinstance(statistic, TFMinMaxTensorStatistic):
        return statistic
    if isinstance(statistic, TFMedianMADTensorStatistic):
        # Using three-sigma approach to estimate min and max
        # Constant factor depends on the distribution form - assuming normal and the factor is 1.4826
        return TFMinMaxTensorStatistic(
            statistic.median_values - 3 * 1.4826230 * statistic.mad_values,
            statistic.median_values + 3 * 1.4826230 * statistic.mad_values,
        )
    if isinstance(statistic, TFPercentileTensorStatistic):
        if len(statistic.percentile_vs_values_dict.keys()) < 2:
            raise ValueError("Cannot create a min-max statistic for less than 2 percentile values")
        min_pct = min(statistic.percentile_vs_values_dict.keys())
        max_pct = max(statistic.percentile_vs_values_dict.keys())
        return TFMinMaxTensorStatistic(
            statistic.percentile_vs_values_dict[min_pct], statistic.percentile_vs_values_dict[max_pct]
        )
    raise ValueError("Unknown TensorStatistic to generate min-max stat from!")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Set, Tuple

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.hardware.config import HWConfig
from nncf.common.quantization.structs import QuantizerConfig
from nncf.experimental.common.tensor_statistics.collectors import TensorReducerBase
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.metatypes.groups import ELEMENTWISE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_WEIGHTS
from nncf.openvino.graph.model_utils import get_start_nodes_for_activation_path_tracing
from nncf.openvino.graph.node_utils import get_weight_channel_axes
from nncf.openvino.graph.transformations.commands import OVConvertInsertionCommand
from nncf.openvino.graph.transformations.commands import OVQuantizerInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.hardware.config import OVHWConfig
from nncf.openvino.quantization.default_quantization import DEFAULT_OV_QUANT_TRAIT_TO_OP_DICT
from nncf.openvino.statistics.collectors import OV_REDUCERS_MAP
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice
from nncf.quantization.algorithms.min_max.backend import MinMaxAlgoBackend
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters
from nncf.quantization.range_estimator import StatisticsType


class OVMinMaxAlgoBackend(MinMaxAlgoBackend):
    @property
    def preserved_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVConvolutionMetatype, om.OVLSTMSequenceMetatype]

    @property
    def mat_mul_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVMatMulMetatype]

    @property
    def post_processing_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVTopKMetatype, om.OVNonMaxSuppressionMetatype]

    @property
    def conv_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVConvolutionMetatype]

    @property
    def elementwise_metatypes(self) -> List[OperatorMetatype]:
        return ELEMENTWISE_OPERATIONS

    @property
    def overflow_fix_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.OVConvolutionMetatype,
            om.OVGroupConvolutionMetatype,
            om.OVConvolutionBackpropDataMetatype,
            om.OVGroupConvolutionBackpropDataMetatype,
            om.OVMatMulMetatype,
        ]

    @property
    def add_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVAddMetatype]

    @property
    def group_conv_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVGroupConvolutionMetatype]

    @property
    def shapeof_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVShapeOfMetatype]

    @property
    def dropout_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def read_variable_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVReadValueMetatype]

    @property
    def scaled_dot_product_attention_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVScaledDotProductAttentionMetatype]

    @property
    def scales_unification_map(self) -> Dict[OperatorMetatype, OperatorMetatype]:
        return {om.OVConcatMetatype: self.overflow_fix_metatypes + self.scaled_dot_product_attention_metatypes}

    @property
    def hw_config(self) -> HWConfig:
        return OVHWConfig

    @property
    def quant_trait_op_dict(self) -> Dict[int, OperatorMetatype]:
        return DEFAULT_OV_QUANT_TRAIT_TO_OP_DICT

    @property
    def reducer_map(self) -> Dict[StatisticsType, TensorReducerBase]:
        return OV_REDUCERS_MAP

    @property
    def supports_inplace_statistics(self) -> bool:
        return True

    @staticmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return get_start_nodes_for_activation_path_tracing(nncf_graph)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_quantizer_insertion_command(
        nncf_graph: NNCFGraph,
        target_point: OVTargetPoint,
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> OVQuantizerInsertionCommand:
        return OVQuantizerInsertionCommand(target_point, parameters)

    @staticmethod
    def create_unified_scales_quantizers_insertion_commands(
        nncf_graph: NNCFGraph,
        target_points: List[OVTargetPoint],
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> List[OVQuantizerInsertionCommand]:
        return [OVQuantizerInsertionCommand(target_point, parameters) for target_point in target_points]

    @staticmethod
    def create_convert_insertion_command(
        target_point: OVTargetPoint,
        parameters: FakeConvertParameters,
    ) -> OVQuantizerInsertionCommand:
        return OVConvertInsertionCommand(target_point, parameters)

    @staticmethod
    def get_target_point_shape(nncf_graph: NNCFGraph, node: NNCFNode, target_point: OVTargetPoint) -> Tuple[int, ...]:
        if target_point.is_weight_target_point():
            return node.layer_attributes.constant_attributes[target_point.port_id]["shape"]

        if target_point.type == TargetType.PRE_LAYER_OPERATION:
            edge = nncf_graph.get_input_edge_by_port_id(node, target_point.port_id)
            return edge.tensor_shape
        elif target_point.type == TargetType.POST_LAYER_OPERATION:
            # NOTE: Assumes that all output edges for the `node` with `output_port_id`
            # equal to `target_point.port_id` should have the same `tensor_shape` value.
            edges = nncf_graph.get_output_edges_by_port_id(node, target_point.port_id)
            return edges[0].tensor_shape

        raise NotImplementedError(f"Unsupported target point type {target_point.type}.")

    @staticmethod
    def get_weight_quantization_axes(node: NNCFNode, target_point: OVTargetPoint, ndims: int) -> Tuple[int]:
        return tuple(get_weight_channel_axes(node))

    @staticmethod
    def get_weight_tensor_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Optional[int]]:
        return node.layer_attributes.get_const_port_ids()

    @staticmethod
    def get_ignored_metatypes(model_type: ModelType, device: TargetDevice) -> List[OperatorMetatype]:
        types = []
        if model_type == ModelType.TRANSFORMER:
            types = [
                om.OVAddMetatype,
                om.OVPowerMetatype,
                om.OVSqueezeMetatype,
                om.OVSubtractMetatype,
                om.OVAvgPoolMetatype,
                om.OVReduceMeanMetatype,
                om.OVReduceL2Metatype,
                om.OVSumMetatype,
                om.OVSquaredDifferenceMetatype,
                om.OVMVNMetatype,
                om.OVGroupNormalizationMetatype,
                om.OVBatchNormMetatype,
                om.OVDivideMetatype,
                om.OVSqrtMetatype,
                om.OVMaximumMetatype,
                # Сomparison operations
                om.OVGreaterEqualMetatype,
                om.OVGreaterMetatype,
                om.OVLessEqualMetatype,
                om.OVLessMetatype,
                om.OVNotEqualMetatype,
                om.OVEqualMetatype,
            ]
            if device != TargetDevice.CPU_SPR:
                types.append(om.OVMultiplyMetatype)
        return types

    @staticmethod
    def get_ignored_names_by_layer_attributes(nncf_graph: NNCFGraph) -> Set[str]:
        ignored_names = set()
        target_nodes = nncf_graph.get_nodes_by_metatypes([om.OVGRUSequenceMetatype])
        for node in target_nodes:
            if (
                isinstance(node.layer_attributes, OVLayerAttributes)
                and node.layer_attributes.input_attributes["linear_before_reset"]
            ):
                ignored_names.add(node.node_name)
        return ignored_names

    def get_weight_nodes(self, nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return [
            node
            for node in nncf_graph.get_all_nodes()
            if isinstance(node.layer_attributes, OVLayerAttributes) and node.metatype in OPERATIONS_WITH_WEIGHTS
        ]

    def is_matmul_with_constant(self, node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return node.metatype in self.mat_mul_metatypes and node.layer_attributes is not None

    @staticmethod
    def get_weight_name(nncf_graph: NNCFGraph, target_point: OVTargetPoint) -> str:
        node = nncf_graph.get_node_by_name(target_point.target_node_name)
        return node.layer_attributes.constant_attributes[target_point.port_id]["name"]

    @staticmethod
    def should_quantize_weight(weight_name: str, quantized_weight_names: Set[str]) -> bool:
        return True



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

import nncf
from nncf import NNCFConfig
from nncf.common.composite_compression import CompositeCompressionAlgorithmBuilder
from nncf.common.composite_compression import CompositeCompressionAlgorithmController
from nncf.config.extractors import extract_algorithm_names
from nncf.tensorflow.algorithm_selector import get_compression_algorithm_builder
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout

TModel = TypeVar("TModel")


class TFCompositeCompressionAlgorithmBuilder(CompositeCompressionAlgorithmBuilder, TFCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)

        algo_names = extract_algorithm_names(config)
        if len(algo_names) < 2:
            raise nncf.ValidationError(
                "Composite algorithm builder must be supplied with a config with more than one "
                "compression algo specified!"
            )
        for algo_name in algo_names:
            algo_builder_cls = get_compression_algorithm_builder(algo_name)
            self._child_builders.append(algo_builder_cls(config, should_init=should_init))

    def _build_controller(self, model: TModel) -> CompositeCompressionAlgorithmController:
        composite_ctrl = CompositeCompressionAlgorithmController(model)
        for builder in self.child_builders:
            composite_ctrl.add(builder.build_controller(model))
        return composite_ctrl

    def get_transformation_layout(self, model: TModel) -> TFTransformationLayout:
        transformations = TFTransformationLayout()
        for builder in self.child_builders:
            transformations.update(builder.get_transformation_layout(model))
        return transformations

    def initialize(self, model: TModel) -> None:
        for builder in self.child_builders:
            builder.initialize(model)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import re
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, TextIO, cast

import nncf
from nncf.common.tensor_statistics.statistics_validator import validate_cache
from nncf.common.utils.backend import BackendType
from nncf.common.utils.os import safe_open
from nncf.tensor import functions as fns
from nncf.tensor.tensor import Tensor
from nncf.tensor.tensor import get_tensor_backend

METADATA_FILE = "statistics_metadata.json"
STATISTICS_FILE_EXTENSION = ".safetensors"


def sanitize_filename(filename: str) -> str:
    """
    Replaces any forbidden characters with an underscore.

    :param filename: Original filename.
    :return: Sanitized filename with no forbidden characters.
    """
    return re.sub(r"[^\w]", "_", filename)


def add_unique_name(name: str, unique_map: Dict[str, List[str]]) -> str:
    """
    Creates an unique name, adds it to a `unique_map` and returns this unique name.

    :param name: The original name.
    :param unique_map: A dictionary mapping names to lists of unique sanitized names.
    :return: A unique name generated by appending a count to the original name.
    """
    # Next number of the same sanitized name
    count = len(unique_map[name]) + 1
    unique_sanitized_name = f"{name}_{count}"
    unique_map[name].append(unique_sanitized_name)
    return unique_sanitized_name


def load_metadata(dir_path: Path) -> Dict[str, Any]:
    """
    Loads the metadata, including the mapping and any other metadata information from the metadata file.

    :param dir_path: The directory where the metadata file is stored.
    :return: A dictionary containing the metadata.
    """
    metadata_file = dir_path / METADATA_FILE
    if metadata_file.exists():
        with safe_open(metadata_file, "r") as f:
            return cast(Dict[str, Any], json.load(f))
    raise nncf.StatisticsCacheError(f"Metadata file does not exist in the following path: {dir_path}")


def save_metadata(metadata: Dict[str, Any], dir_path: Path) -> None:
    """
    Saves metadata to a file in the specified directory.

    :param metadata: Dictionary containing metadata and mapping.
    :param dir_path: Path to the directory where the metadata file will be saved.
    """
    metadata_file = dir_path / METADATA_FILE
    with safe_open(metadata_file, "w") as f:
        json.dump(metadata, cast(TextIO, f), indent=4)


def load_statistics(dir_path: Path, backend: BackendType) -> Dict[str, Dict[str, Tensor]]:
    """
    Loads statistics from a directory.

    :param dir_path: The path to the directory from which to load the statistics.
    :param backend: Backend type to determine the tensor backend.
    :return: Statistics.
    """
    metadata = load_metadata(dir_path)
    try:
        validate_cache(metadata, dir_path, backend)
        statistics: Dict[str, Dict[str, Tensor]] = {}
        mapping = metadata.get("mapping", {})
        tensor_backend = get_tensor_backend(backend)
        for file_name, original_name in mapping.items():
            statistics_file = dir_path / file_name
            statistics[original_name] = fns.io.load_file(statistics_file, backend=tensor_backend)  # no device support
        return statistics
    except Exception as e:
        raise nncf.StatisticsCacheError(str(e))


def dump_statistics(
    statistics: Dict[str, Dict[str, Tensor]],
    dir_path: Path,
    backend: BackendType,
    additional_metadata: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Saves statistics and metadata to a directory.

    Metadata is stored in a JSON file named "statistics_metadata.json".
    Statistics are stored in individual files with sanitized and unique filenames to prevent collisions.

    Metadata Format:
    The metadata file must have a mapping of saved filenames to the original names and backend type.
    {
        "mapping": {
            "saved_file_name_1": "original_name_1",
            "saved_file_name_2": "original_name_2",
            ...
        },
        "backend": "backend_type",
        ... (additional metadata fields)
    }

    :param statistics: A dictionary with statistic names as keys and the statistic data as values.
    :param dir_path: The path to the directory where the statistics will be dumped.
    :param backend: Backend type to save in metadata.
    :param additional_metadata: A dictionary containing any additional metadata to be saved with the mapping.
    """
    dir_path.mkdir(parents=True, exist_ok=True)
    metadata: Dict[str, Any] = {"mapping": {}, "backend": backend.value}
    unique_map: Dict[str, List[str]] = defaultdict(list)
    for original_name, statistics_value in statistics.items():
        sanitized_name = sanitize_filename(original_name)
        unique_sanitized_name = add_unique_name(sanitized_name, unique_map) + STATISTICS_FILE_EXTENSION

        file_path = dir_path / unique_sanitized_name

        # Update the mapping
        metadata["mapping"][unique_sanitized_name] = original_name
        fns.io.save_file(statistics_value, file_path)

    if additional_metadata:
        metadata |= additional_metadata

    save_metadata(metadata, dir_path)
